import time
import inspect
import traceback
import multiprocessing as mp
import io

class PipeWrapper(object):
    def __init__(self, pipe=None, log_filename='', cmd_event=None, **kwargs):
        self.pipe = pipe
        self.log_filename = log_filename
        self.cmd_event = cmd_event

    def log_error(self, err, mode='a'):
        if self.log_filename != '':
            traceback.print_exc(None, err)
            with open(self.log_filename, mode) as fp:
                err.seek(0)
                fp.write(err.read())

    def log_str(self, s, mode="a", newline=True):
        if self.log_filename != '':
            if newline and not s.endswith("\n"):
                s += "\n"
            with open(self.log_filename, mode) as fp:
                fp.write(s)


class FuncProxy(object):
    '''
    Interface for calling functions in remote processes.
    '''
    def __init__(self, name, pipe, event, lock, log_filename=''):
        '''
        Constructor for FuncProxy

        Parameters
        ----------
        name : string
            Name of remote function to call
        pipe : mp.Pipe instance
            multiprocessing pipe through which to send data (function name, arguments) and receive the result
        event : mp.Event instance
            A flag to set which is multiprocessing-compatible (visible to both the current and the remote processes)

        Returns
        -------
        FuncProxy instance
        '''
        self.pipe = pipe
        self.name = name
        self.event = event
        self.lock = lock
        self.log_filename = log_filename

    def log_error(self, err, mode='a'):
        if self.log_filename != '':
            traceback.print_exc(None, err)
            with open(self.log_filename, mode) as fp:
                err.seek(0)
                fp.write(err.read())

    def log_str(self, s, mode="a", newline=True):
        if self.log_filename != '':
            if newline and not s.endswith("\n"):
                s += "\n"
            with open(self.log_filename, mode) as fp:
                fp.write(s)

    def __call__(self, *args, **kwargs):
        '''
        Return the result of the remote function call

        Parameters
        ----------
        *args, **kwargs : positional arguments, keyword arguments
            To be passed to the remote function associated when the object was created

        Returns
        -------
        function result
        '''
        self.lock.acquire()
        self.log_str('lock acquired')

        # block until the remote process unsets the event. Acts as a lock
        n_ticks = 0
        while self.event.is_set() and n_ticks < 100:
            n_ticks += 1
            time.sleep(0.1)

        if n_ticks >= 100:
            raise Exception("Out of sync!")

        # print("FuncProxy.__call__", self.name, args, kwargs)

        self.pipe.send((self.name, args, kwargs))
        self.event.set()
        # resp = self.pipe.recv()
        if self.pipe.poll(10):
            resp = self.pipe.recv()
        else:
            raise Exception("FuncProxy: remote object failed to respond")

        self.lock.release()        
        self.log_str('lock released') 
        return resp


class ObjProxy(PipeWrapper):
    def __init__(self, target_class, *args, **kwargs):
        self.target_class = target_class
        is_instance_method = lambda n: inspect.isfunction(getattr(self.target_class, n))
        self.methods = set(filter(is_instance_method, dir(self.target_class)))

        super().__init__(*args, **kwargs)
        self.lock = mp.Lock()

    def __getattr__(self, attr):
        self.log_str("remotely getting attribute: %s\n" % attr)

        methods = object.__getattribute__(self, "methods")
        pipe = object.__getattribute__(self, "pipe")
        cmd_event = object.__getattribute__(self, "cmd_event")
        lock = object.__getattribute__(self, "lock")

        if attr in methods:
            self.log_str("returning function proxy for %s" % attr)
            return FuncProxy(attr, pipe, cmd_event, lock, log_filename=self.log_filename)
        else:
            self.log_str("sending __getattribute__ over pipe")
            fn_getattr = FuncProxy("__getattribute__", pipe, cmd_event, lock, log_filename=self.log_filename)
            return fn_getattr(attr)

    def set(self, attr, value):
        self.log_str("ObjProxy.setattr")
        pipe = object.__getattribute__(self, "pipe")
        cmd_event = object.__getattribute__(self, "cmd_event")        
        lock = object.__getattribute__(self, "lock")
        setattr_fn = FuncProxy('__setattr__', pipe, cmd_event, lock, log_filename=self.log_filename)

        setattr_fn(attr, value)
        self.log_str("Finished setting remote attr %s to %s" % (str(attr), str(value)))

    def terminate(self):
        self.pipe.send(None)


class DataPipe(PipeWrapper):
    pass

def call_from_remote(x):
    return x

def call_from_parent(x):
    return x

class RPCProcess(mp.Process):
    """mp.Process which implements remote procedure call (RPC) through a mp.Pipe object"""
    def __init__(self, target_class=object, target_kwargs=dict(), log_filename='', **kwargs):
        super().__init__()
        self.cmd_pipe = None
        self.data_pipe = None
        self.log_filename = log_filename

        self.target = None
        self.target_class = target_class
        self.target_kwargs = target_kwargs

        self.cmd_event = mp.Event()
        self.status = mp.Value('b', 1) # mp boolean used for terminating the remote process

        self.target_proxy = None
        self.data_proxy = None

    def __getattr__(self, attr):
        """ Redirect attribute access to the target object if the 
        attribute can't be found in the process wrapper """
        try:
            if self.target_proxy is not None and self.status.value > 0:
                try:
                    return getattr(self.target_proxy, attr)
                except:
                    raise AttributeError("RPCProcess: could not forward getattr %s to target of type %s" % (attr, self.target_class))
            else:
                raise AttributeError("RPCProcess: target proxy not initialized")
        except:
            raise AttributeError("Could not get RPCProcess attribute: %s" % attr)

    def log_error(self, err, mode='a'):
        if self.log_filename != '':
            with open(self.log_filename, mode) as fp:
                traceback.print_exc(file=fp)
                fp.write(str(err))

    def log_str(self, s, mode="a", newline=True):
        if self.log_filename != '':
            if newline and not s.endswith("\n"):
                s += "\n"
            with open(self.log_filename, mode) as fp:
                fp.write(s)

    @call_from_remote
    def target_constr(self):
        try:
            self.target = self.target_class(**self.target_kwargs)
            if hasattr(self.target, 'start'):
                self.target.start()
        except Exception as e:
            print("RPCProcess.target_constr: unable to start source!")
            print(e)

            import io
            err = io.StringIO()
            self.log_error(err, mode='a')
            err.seek(0)

            self.status.value = -1

    @call_from_remote
    def target_destr(self, ret_status, msg):
        pass

    @call_from_parent
    def start(self):
        self.cmd_pipe, pipe_end2 = mp.Pipe()
        self.data_pipe, data_pipe2 = mp.Pipe()

        self.target_proxy = ObjProxy(self.target_class, pipe=pipe_end2, 
            log_filename=self.log_filename, cmd_event=self.cmd_event)
        self.data_proxy = DataPipe(data_pipe2)
        super().start()
        return self.target_proxy, self.data_proxy

    def check_run_condition(self):
        return self.status.value > 0

    def is_enabled(self):
        return self.status.value > 0

    @call_from_parent
    def stop(self):
        self.status.value = -1

    def __del__(self):
        '''Stop the process when the object is destructed'''
        if self.status.value > 0:
            #self.stop() <- currently causing issues with task_wrapper. Somewhere the status is not being set properly after the task ends...
            self.status.value = -1

    def is_cmd_present(self):
        return self.cmd_event.is_set()

    def clear_cmd(self):
        self.log_str("clearing")
        self.cmd_event.clear()

    @call_from_remote
    def loop_task(self):
        time.sleep(0.01)

    @call_from_remote
    def run(self):
        self.log_str("RPCProcess.run")

        self.target_constr()

        try:
            while self.is_enabled():
                if not self.check_run_condition():
                    self.log_str("The target's termination condition was reached")
                    break

                if self.is_cmd_present():
                    self.proc_rpc_command()
                    self.clear_cmd()

                self.loop_task()

            self.log_str("RPCProcess.run: end of while loop")
            ret_status, msg = 0, ''
        except KeyboardInterrupt:
            ret_status, msg = 1, 'KeyboardInterrupt'
        except:
            import traceback
            traceback.print_exc()
            err = io.StringIO()
            self.log_error(err, mode='a')
            err.seek(0)
            ret_status, msg = 1, err.read()
        finally:
            self.target_destr(ret_status, msg)
            self.status.value = -1

    @call_from_remote
    def proc_rpc_command(self):
        cmd = self.cmd_pipe.recv()
        self.log_str("Received command: %s" % str(cmd))

        if cmd is None:
            self.stop()
            return 

        try:
            fn_name, cmd_args, cmd_kwargs = cmd
            fn = getattr(self.target, fn_name)
            self.log_str("Function: " + str(fn))

            fn_output = fn(*cmd_args, **cmd_kwargs)
            self.cmd_pipe.send(fn_output)

            self.log_str('Done with command: %s, output=%s\n\n' % (fn_name, fn_output))
        except Exception as e:
            err = io.StringIO()
            self.log_error(err, mode='a')

            self.cmd_pipe.send(e)




'''
Client-side code that uses cbpy to configure and receive neural data from the 
Blackrock Neural Signal Processor (NSP) (or nPlay).
'''

import sys
import time
from collections import namedtuple
try:
    from cerebus import cbpy
except ImportError:
    import warnings
    warnings.warn("Unable to import cerebus library. Check if is installed if using the Blackrock NeuroPort system")

SpikeEventData = namedtuple("SpikeEventData",
                            ["chan", "unit", "ts", "arrival_ts"])
ContinuousData = namedtuple("ContinuousData", 
                            ["chan", "samples", "arrival_ts"])

class Connection(object):
    '''
    A wrapper around a UDP socket which sends the Blackrock NeuroPort system commands and 
    receives data. Must run in a separte process (e.g., through `riglib.source`) 
    if you want to use it as part of a task (e.g., BMI control)
    '''
    debug = False
    def __init__(self):
        self.parameters = dict()
        self.parameters['inst-addr']   = '192.168.137.128'
        self.parameters['inst-port']   = 51001
        self.parameters['client-port'] = 51002

        self.channel_offset = 0  # used to be 4 -- some old bug with nPlay
        print('Using cbpy channel offset of:', self.channel_offset)

        if sys.platform == 'darwin':  # OS X
            print('Using OS X settings for cbpy')
            self.parameters['client-addr'] = '255.255.255.255'
        else:  # linux
            print('Using linux settings for cbpy')
            self.parameters['client-addr'] = '192.168.137.255'
            self.parameters['receive-buffer-size'] = 8388608

        self._init = False

        if self.debug:
            self.nsamp_recv = 0
            self.nsamp_last_print = 0
    
    def connect(self):
        '''Open the interface to the NSP (or nPlay).'''

        print('calling cbpy.open in cerelink.connect()')
        # try:
        #     result, return_dict = cbpy.open(connection='default')
        #     time.sleep(3)
        # except:
        result, return_dict = cbpy.open(connection='default', parameter=self.parameters)
        time.sleep(3)

        print('cbpy.open result:', result)
        print('cbpy.open return_dict:', return_dict)
        if return_dict['connection'] != 'Master':
            raise Exception
        print('')
        
        # return_dict = cbpy.open('default', self.parameters)  # old cbpy
        
        self._init = True
        
    def select_channels(self, channels):
        '''Sets the channels on which to receive event/continuous data.

        Parameters
        ----------
        channels : array_like
            A sorted list of channels on which you want to receive data.
        '''
        
        if not self._init:
            raise ValueError("Please open the interface to Central/nPlay first.")

        buffer_parameter = {'absolute': True}  # want absolute timestamps

        # ability to select desired channels not yet implemented in cbpy        
        # range_parameter = dict()
        # range_parameter['begin_channel'] = channels[0]
        # range_parameter['end_channel']   = channels[-1]

        print('calling cbpy.trial_config in cerelink.select_channels()')
        result, reset = cbpy.trial_config(buffer_parameter=buffer_parameter)
        print('cbpy.trial_config result:', result)
        print('cbpy.trial_config reset:', reset)
        print('')
    
    def start_data(self):
        '''Start the buffering of data.'''
        
        if not self._init:
            raise ValueError("Please open the interface to Central/nPlay first.")

        self.streaming = True

    def stop_data(self):
        '''Stop the buffering of data.'''
        
        if not self._init:
            raise ValueError("Please open the interface to Central/nPlay first.")

        print('calling cbpy.trial_config in cerelink.stop()')
        result, reset = cbpy.trial_config(reset=False)
        print('cbpy.trial_config result:', result)
        print('cbpy.trial_config reset:', reset)
        print('')

        self.streaming = False

    def disconnect(self):
        '''Close the interface to the NSP (or nPlay).'''
        
        if not self._init:
            raise ValueError("Please open the interface to Central/nPlay first.")
        
        print('calling cbpy.close in cerelink.disconnect()')
        result = cbpy.close()
        print('result:', result)
        print('')

        self._init = False
    
    def __del__(self):
        self.disconnect()

    def get_event_data(self):
        '''A generator that yields spike event data.'''

        sleep_time = 0

        while self.streaming:

            result, trial = cbpy.trial_event(reset=True)  # TODO -- check if result = 0?
            arrival_ts = time.time()

            for list_ in trial:
                chan = list_[0]
                for unit, unit_ts in enumerate(list_[1]['timestamps']):
                    for ts in unit_ts:
                        # blackrock unit numbers are 0-based where zero is unsorted unit
                        if unit == 0:
                            # Unsorted units are unit 10 (j)
                            un = 10
                        else:
                            un = unit
                        yield SpikeEventData(chan=chan-self.channel_offset, unit=un, ts=ts, arrival_ts=arrival_ts)

            time.sleep(sleep_time)


    def get_continuous_data(self):
        '''A generator that yields continuous data.'''

        sleep_time = 0

        while self.streaming:
            result, trial = cbpy.trial_continuous(reset=True)
            arrival_ts = time.time()

            for list_ in trial:

                if self.debug:
                    chan = list_[0]
                    samples = list_[1]
                    if chan == 8:
                        self.nsamp_recv += len(samples)
                        if self.nsamp_recv > self.nsamp_last_print + 2000:
                            print("cerelink.py: # received =", self.nsamp_recv)
                            self.nsamp_last_print = self.nsamp_recv

                yield ContinuousData(chan=list_[0],
                                     samples=list_[1],
                                     arrival_ts=arrival_ts)

            time.sleep(sleep_time)




import numpy as np

def dist(a, b):
    '''Calculate the Euclidean distance between vectors a and b.'''

    return np.sqrt(sum((a[i] - b[i])**2 for i in range(len(a))))

def norm_vec(x, eps=1e-9):
    return x / (np.linalg.norm(x) + eps)

def bound(x, x_min, x_max):
    if x_min > x_max:
        raise Exception('Error in bound function: min bound is greater than max bound!')
    
    x_bounded = x
    if not np.isnan(x_min):
        x_bounded = max(x_bounded, x_min)
    if not np.isnan(x_max):
        x_bounded = min(x_bounded, x_max)

    return x_bounded



'''
Features for a touch sensor on the neurosync arduino
'''
from riglib.experiment import traits
from riglib import touch_data

########################################################################################################
# Touch sensor datasources
########################################################################################################

class TouchDataFeature(traits.HasTraits):
    '''
    Enable reading of data from touch sensor
    '''

    def init(self):
        '''
        Secondary init function. See riglib.experiment.Experiment.init()
        Prior to starting the task, this 'init' sets up the DataSource for interacting with the 
        kinarm system and registers the source with the SinkRegister so that the data gets saved 
        to file as it is collected.
        '''
        from riglib import source
        System  = touch_data.TouchData
        self.touch_data = source.DataSource(System)
        from riglib import sink
        sink_manager = sink.SinkManager.get_instance()
        sink_manager.register(self.touch_data)
        super(TouchDataFeature, self).init()
    
    @property
    def source_class(self):
        '''
        Specify the source class as a function
        '''
        from riglib import touch_data
        return touch_data.TouchData()

    def run(self):
        '''
        Code to execute immediately prior to the beginning of the task FSM executing, or after the FSM has finished running. 
        See riglib.experiment.Experiment.run(). This 'run' method starts the motiontracker source prior to starting the experiment's 
        main thread/process, and handle any errors by stopping the source
        '''
        self.touch_data.start()
        try:
            super(TouchDataFeature, self).run()
        finally:
            self.touch_data.stop()
    
    def join(self):
        '''
        See riglib.experiment.Experiment.join(). Re-join the 'motiondata' source process before cleaning up the experiment thread
        '''
        #self.touch_data.join()
        super(TouchDataFeature, self).join()
    
    def _start_None(self):
        '''
        Code to run before the 'None' state starts (i.e., the task stops)
        '''
        self.touch_data.stop()
        super(TouchDataFeature, self)._start_None()




from db.tracker import models
from db import dbfunctions as dbfn
from db.tracker import trainbmi
import numpy as np
from .kfdecoder import KalmanFilter, KFDecoder
from . import train
import pickle
import re
import tables
import os

########## MAIN DECODER MANIPULATION METHODS #################

def add_rm_units(task_entry_id, units, add_or_rm, flag_added_for_adaptation, name_suffix='', decoder_entry_id=None, **kwargs):
    '''
    Summary: Method to add or remove units from KF decoder. 
        Takes in task_entry_id number or decoder_entry_id to get decoder
        Removes or adds units to decoder
            If adds, sets decoder weights to random entries


    Input param: task_entry_id: Decoder = dbfn.TaskEntry(task_entry_id).get_decoders_trained_in_block()
    Input param: units: list of units to add or remove
    Input param: add_or_rm: 'add' or 'rm' 
    Input param: flag_added_for_adaptation: whether or not to flag newly added units as adapting inds
    Input param: name_suffix: new decoder suffix. If empty, will append 'add_or_rm_units_len(units)'
    Input param: decoder_entry_id: used if more than 1 decoder training on block

    '''
    if 'decoder_path' in kwargs:
        kfdec = pickle.load(open(kwargs['decoder_path']))
    else:
        kfdec = get_decoder_corr(task_entry_id, decoder_entry_id)

    if add_or_rm is 'add':
        kfdec_new , n_new_units = add_units(kfdec, units)
        
        # Only Adapt new units: 
        if flag_added_for_adaptation:
            kfdec_new.adapting_neural_inds = np.arange(len(kfdec_new.units)-len(units), len(kfdec_new.units))

        save_new_dec(task_entry_id, kfdec_new, name_suffix+'_add_'+str(n_new_units)+'_units')

    elif add_or_rm is 'rm':
        orig_units = kfdec.units
        inds_to_keep = proc_units(kfdec, units, 'remove')
        if len(orig_units) == len(inds_to_keep):
            print(' units cannot be removed since theyre not in original decoder', orig_units)
        else:
            dec_new = return_proc_units_decoder(kfdec, inds_to_keep)
            save_new_dec(task_entry_id, dec_new, name_suffix+'_rm_'+str(len(units))+'_units')

def flag_adapting_inds_for_CLDA(task_entry_id, state_names_to_adapt=None, units_to_adapt = None, decoder_entry_id=None):
    
    decoder = get_decoder_corr(task_entry_id, decoder_entry_id)
    state_adapting_inds = []
    for s in state_names:
        ix = np.nonzero(decoder.states == s)[0]
        assert len(ix) == 0
        state_adapting_inds.append(int(ix))
    decoder.adapting_state_inds = np.array(adapting_inds)

    neural_adapting_inds = []
    for u in units_to_adapt:
        uix = np.nonzero(np.logical_and(decoder.units[:, 0]== u[0], decoder.units[:, 1]== u[1]))[0]
        neural_adapting_inds.append(int(uix))
    decoder.adapting_neural_inds = np.array(neural_adapting_inds)
    save_new_dec(task_entry_id, decoder, '_adapt_only_'+str(len(units_to_adapt))+'_units_'+str(len(state_names_to_adapt))+'_states')

def zscore_units(task_entry_id, calc_zscore_from_te, pos_key = 'cursor', decoder_entry_id=None, 
    training_method=train.train_KFDecoder, retrain_flag = False, **kwargs):
    '''
    Summary: Method to be able to 'convert' a trained decoder (that uses zscoring) to one that uses z-scored from another session
         (e.g. you train a decoder from VFB, but you want to zscore unit according to a passive / still session earlier). You 
         would use the task_entry_id that was used to train the decoder OR entry that used the decoder. Then 'calc_zscore_from_te'
         is the task entry ID used to compute the z-scored units. You can either retrain the decoder iwth the new z-scored units, or not

    Input param: task_entry_id:
    Input param: decoder_entry_id:
    Input param: calc_zscore_from_te:
    Output param: 
    '''
    if 'decoder_path' in kwargs:
        decoder = pickle.load(open(kwargs['decoder_path']))
    else:
        decoder = get_decoder_corr(task_entry_id, decoder_entry_id, get_dec_used=False)

    assert (hasattr(decoder, 'zscore') and decoder.zscore is True)," Cannot update mFR /sdFR of decoder that was not trained as zscored decoder. Retrain!"

    # Init mFR / sdFR
    if 'hdf_path' in kwargs:
        hdf = tables.openFile(kwargs['hdf_path'])
    else:
        hdf = dbfn.TaskEntry(calc_zscore_from_te).hdf
    
    # Get HDF update rate from hdf file. 
    try:
        hdf_update_rate = np.round(np.mean(hdf.root.task[:]['loop_time'])*1000.)/1000.
    except:
        from config import config
        if config.recording_system == 'blackrock':
            hdf_update_rate = .05;
        elif config.recording_system == 'plexon':
            hdf_update_rate = 1/60.

    spk_counts = hdf.root.task[:]['spike_counts'][:, :, 0]
    
    # Make sure not repeated entries:
    sum_spk_counts = np.sum(spk_counts, axis=1)
    ix = np.nonzero(sum_spk_counts)[0][0]
    sample = 1+ sum_spk_counts[ix:ix+6] - sum_spk_counts[ix]
    assert np.sum(sample) != 6

    decoder_update_rate = decoder.binlen
    bin_spks, _ = bin_(None, spk_counts.T, hdf_update_rate, decoder_update_rate, only_neural=True)
    mFR = np.squeeze(np.mean(bin_spks, axis=1))
    sdFR = np.std(bin_spks, axis=1)
    kwargs2 = dict(mFR=mFR, sdFR=sdFR)

    #Retrain decoder w/ new zscoring:
    if 'te_id' in kwargs:
        training_id = kwargs['te_id']
    else:
        training_id = decoder.te_id
    

    if retrain_flag:
        raise NotImplementedError("Need to test retraining with real data")
        saved_files = models.DataFile.objects.filter(entry_id=training_id)
        files = {}
        for fl in saved_files:
            files[fl.system.name] = fl.get_path()
        import bmilist
        decoder = training_method(files, decoder.extractor_cls, decoder.extractor_kwargs, bmilist.kin_extractors[''], decoder.ssm, 
            decoder.units, update_rate=decoder.binlen, tslice=decoder.tslice, pos_key=pos_key, zscore=True, **kwargs2)
        suffx = '_zscore_set_from_'+str(calc_zscore_from_te)+'_retrained'
    else:
        decoder.mFR = 0.
        decoder.sdFR = 1.
        decoder.init_zscore(mFR, sdFR)
        decoder.mFR = mFR
        decoder.sdFR = sdFR

        suffx = '_zscore_set_from_'+str(calc_zscore_from_te)

    if task_entry_id is not None:
        save_new_dec(task_entry_id, decoder, suffx)
    else:
        return decoder, suffx

def zscore_features(task_entry_id, calc_zscore_from_te, decoder_entry_id=None, kin_source = 'cursor', **kwargs):
    '''
    Summary: Method to be able to 'convert' a trained decoder (that uses zscoring) to one that uses z-scored from another session
         (e.g. you train a decoder from VFB, but you want to zscore unit according to a passive / still session earlier). You 
         would use the task_entry_id that was used to train the decoder OR entry that used the decoder. Then 'calc_zscore_from_te'
         is the task entry ID used to compute the z-scored features. Unlike zscore_units, this recomputes the features using the 
         feature extractor and extractor_kwargs stored in the decoder.

    Input param: task_entry_id:
    Input param: decoder_entry_id:
    Input param: calc_zscore_from_te:
    Output param: 
    '''
    if 'decoder_path' in kwargs:
        decoder = pickle.load(open(kwargs['decoder_path']))
    else:
        decoder = get_decoder_corr(task_entry_id, decoder_entry_id, get_dec_used=False)

    # assert (hasattr(decoder, 'zscore') and decoder.zscore is True)," Cannot update mFR /sdFR of decoder that was not trained as zscored decoder. Retrain!"

    # Extract new features
    te = models.TaskEntry.objects.get(id=calc_zscore_from_te)
    te_json = te.to_json()
    files = te.get_data_files_dict_absolute()
    binlen = decoder.binlen
    units = decoder.units
    extractor_cls = decoder.extractor_cls
    extractor_kwargs = decoder.extractor_kwargs

    # Load the recording to get the length
    recording_sys_make = models.KeyValueStore.get('recording_sys')
    if recording_sys_make == 'plexon':
        from plexon import plexfile # keep this import here so that only plexon rigs need the plexfile module installed
        plexon = models.System.objects.get(name='plexon')
        df = models.DataFile.objects.get(entry=te.id, system=plexon)
        plx = plexfile.openFile(df.get_path().encode('utf-8'), load=False)
        length = plx.length
    elif recording_sys_make == 'ecube':
        sys = models.System.objects.get(name='ecube')
        df = models.DataFile.objects.get(entry=te.id, system=sys)
        filepath = df.get_path()
        from riglib.ecube import parse_file
        info = parse_file(str(df.get_path()))
        length = info.length
    else:
        ValueError('Unrecognized recording_system!')

    tslice = [0, length]
    strobe_rate = te.task_params['fps']

    neural_features, units, extractor_kwargs = train.get_neural_features(files, binlen, extractor_cls.extract_from_file,
        extractor_kwargs, tslice=tslice, units=units, source=kin_source, strobe_rate=strobe_rate)
    mFR = np.squeeze(np.mean(neural_features, axis=0))
    sdFR = np.std(neural_features, axis=0)

    # Update decoder w/ new zscoring:   
    decoder.mFR = 0.
    decoder.sdFR = 1.
    decoder.init_zscore(mFR, sdFR)
    decoder.mFR = mFR
    decoder.sdFR = sdFR

    # Save it as a new decoder
    suffx = '_zscore_set_from_'+str(calc_zscore_from_te)

    if task_entry_id is not None:
        save_new_dec(task_entry_id, decoder, suffx)
    else:
        return decoder, suffx



def adj_state_noise(task_entry_id, decoder_entry_id, new_w):
    decoder = get_decoder_corr(task_entry_id, decoder_entry_id, return_used_te=True)
    W = np.diag(decoder.filt.W)
    wix = np.nonzero(W)[0]
    W[wix] = new_w
    W_new = np.diag(W)
    decoder.filt.W = W_new
    save_new_dec(task_entry_id, decoder, '_W_new_'+str(new_w))    

########################## HELPER DECODER MANIPULATION METHODS #################################

def get_decoder_corr(task_entry_id, decoder_entry_id, get_dec_used=True):
    '''
    Summary: get KF decoder either from entry that has trained the decoder (if this, need decoder_entry_id if > 1 decoder), 
        or decoder that was used during task_entry_id
    Input param: task_entry_id: dbname task entry ID
    Input param: decoder_entry_id: decoder entry id: (models.Decoder.objects.get(entry=entry))
    Output param: KF Decoder
    '''
    ld = True
    if get_dec_used is False:
        decoder_entries = dbfn.TaskEntry(task_entry_id).get_decoders_trained_in_block()
        if type(decoder_entries) is models.Decoder:
            decoder = decoder_entries
            ld = False
        else: # list of decoders. Search for the right one. 
            try:
                dec_ids = [de.pk for de in decoder_entries]
                _ix = np.where(np.isin(dec_ids, decoder_entry_id))[0][0]
                decoder = decoder_entries[_ix]
                ld = False
            except:
                if decoder_entry_id is None:
                    print('Too many decoder entries trained from this TE, specify decoder_entry_id')
                else:
                    print('Too many decoder entries trained from this TE, no match to decoder_entry_id %d'%decoder_entry_id)
    if ld is False:
        kfdec = decoder.load()            
    else:
        try:
            kfdec = dbfn.TaskEntry(task_entry_id).decoder
            print('Loading decoder USED in task %s'%dbfn.TaskEntry(task_entry_id).task)
        except:
            raise Exception('Cannot load decoder from TE%d'%task_entry_id)
    return kfdec

def add_units(kfdec, units):
    '''
    Add units to KFDecoder, e.g. to account for appearance of new cells 
    on a particular day, will need to do CLDA to fit new deocder weight
    
    Parameters: 
    units: string or np.ndarray of shape (N, 2) of units to ADD to current decoder
    '''
    units_curr = kfdec.units
    new_units = proc_units(kfdec, units, 'to_int')

    keep_ix = []
    for r, r_un in enumerate(new_units):
        if len(np.nonzero(np.all(r_un==units_curr, axis=1))[0]) > 0: 
            print('not adding unit ', r_un, ' -- already in decoder')
        else:
            keep_ix.append(r)

    new_units = np.array(new_units)[keep_ix, :]
    units = np.vstack((units_curr, new_units))
    n_states = kfdec.filt.C.shape[1]
    n_features = len(units)

    C = np.vstack(( kfdec.filt.C, 1e-3*np.random.randn(len(new_units), kfdec.ssm.n_states)))
    Q = np.eye( len(units), len(units) )
    Q[np.ix_(np.arange(len(units_curr)), np.arange(len(units_curr)))] = kfdec.filt.Q
    Q_inv = np.linalg.inv(Q)

    if isinstance(kfdec.mFR, np.ndarray):
        mFR = np.hstack(( kfdec.mFR, np.zeros((len(new_units))) ))
        sdFR = np.hstack(( kfdec.sdFR, np.ones((len(new_units))) ))
    else:
        mFR = kfdec.mFR
        sdFR = kfdec.sdFR

    filt = KalmanFilter(A=kfdec.filt.A, W=kfdec.filt.W, C=C, Q=Q, is_stochastic=kfdec.filt.is_stochastic)
    C_xpose_Q_inv = C.T * Q_inv
    C_xpose_Q_inv_C = C.T * Q_inv * C
    filt.C_xpose_Q_inv = C_xpose_Q_inv
    filt.C_xpose_Q_inv_C = C_xpose_Q_inv_C        

    filt.R = kfdec.filt.R
    ix = np.random.permutation(n_features)[:len(new_units)]
    filt.S = np.vstack(( kfdec.filt.S, kfdec.filt.S[ix, :]))
    filt.T = Q + filt.S * filt.S.T
    filt.ESS = kfdec.filt.ESS

    decoder = KFDecoder(filt, units, kfdec.ssm, mFR=mFR, sdFR=sdFR, binlen=kfdec.binlen, tslice=kfdec.tslice)
    decoder.n_features = units.shape[0]
    decoder.units = units
    decoder.extractor_cls = kfdec.extractor_cls
    decoder.extractor_kwargs = kfdec.extractor_kwargs
    try:
        CE = kfdec.corresp_encoder
        CE.C = np.vstack((CE.C, 3*np.random.randn(len(new_units), CE.C.shape[1])))
        Q = .1*np.eye(len(units))
        Q[:len(units_curr), :len(units_curr)] = CE.Q
        CE.Q = Q
        CE.n_features = len(units)
        decoder.corresp_encoder = CE
        print('adjusted corresp_encoder too!')
    except:
        pass
    decoder.extractor_kwargs['units'] = units
    return decoder, len(keep_ix)

def proc_units(kfdec, units, mode):
    '''
    Parse list of units indices to keep from string or np.ndarray of shape (N, 2)
    Inputs: 
        units -- 
        mode -- can be 'keep' or 'remove' or 'to_int'. Tells function what to do with the units
    '''

    if isinstance(units[0], str):
        # convert to array
        if isinstance(units, str):
            units = units.split(', ')

        units_lut = dict(a=1, b=2, c=3, d=4, e=5, f=6, g=7, h=8, i=9, j=10, k=11)
        units_int = []
        for u in units:
            ch = int(re.match('(\d+)([a-k])', u).group(1))
            unit_ind = re.match('(\d+)([a-k])', u).group(2)
            # import pdb; pdb.set_trace()
            units_int.append((ch, units_lut[unit_ind]))

        units = units_int
    
    if mode == 'to_int':
        return units

    inds_to_keep = []
    new_units = list(map(tuple, units))
    for k, old_unit in enumerate(kfdec.units):
        if mode == 'keep':
            if tuple(old_unit) in new_units:
                inds_to_keep.append(k)
        elif mode == 'remove':
            if tuple(old_unit) not in new_units:
                inds_to_keep.append(k)
    return inds_to_keep

def return_proc_units_decoder(kfdec, inds_to_keep):
    A = kfdec.filt.A
    W = kfdec.filt.W
    C = kfdec.filt.C
    Q = kfdec.filt.Q
    print('Indices to keep: ', inds_to_keep)
    C = C[inds_to_keep, :]
    Q = Q[np.ix_(inds_to_keep, inds_to_keep)]
    Q_inv = np.linalg.inv(Q)

    if isinstance(kfdec.mFR, np.matrix):
        mFR = np.squeeze(np.array(kfdec.mFR))[inds_to_keep]
        sdFR = np.squeeze(np.array(kfdec.sdFR))[inds_to_keep]
        
    elif isinstance(kfdec.mFR, np.ndarray):
        mFR = kfdec.mFR[inds_to_keep]
        sdFR = kfdec.mFR[inds_to_keep]
    else:
        mFR = kfdec.mFR
        sdFR = kfdec.sdFR

    filt = KalmanFilter(A=A, W=W, C=C, Q=Q, is_stochastic=kfdec.filt.is_stochastic)
    C_xpose_Q_inv = C.T * Q_inv
    C_xpose_Q_inv_C = C.T * Q_inv * C
    filt.C_xpose_Q_inv = C_xpose_Q_inv
    filt.C_xpose_Q_inv_C = C_xpose_Q_inv_C        

    units = kfdec.units[inds_to_keep]

    filt.R = kfdec.filt.R
    filt.S = kfdec.filt.S[inds_to_keep, :]
    filt.T = kfdec.filt.T[np.ix_(inds_to_keep, inds_to_keep)]
    filt.ESS = kfdec.filt.ESS

    decoder = KFDecoder(filt, units, kfdec.ssm, mFR=mFR, sdFR=sdFR, binlen=kfdec.binlen, tslice=kfdec.tslice)

    decoder.n_features = units.shape[0]
    decoder.units = units
    decoder.extractor_cls = kfdec.extractor_cls
    decoder.extractor_kwargs = kfdec.extractor_kwargs
    decoder.extractor_kwargs['units'] = units
    try:
        CE = kfdec.corresp_encoder
        CE.C = CE.C[inds_to_keep, :]
        CE.Q = CE.Q[np.ix_(inds_to_keep, inds_to_keep)]
        CE.n_features = len(units)
        decoder.corresp_encoder = CE
        print('adjusted corresp_encoder too!')
    except:
        pass

    return decoder

def save_new_dec(task_entry_id, dec_obj, suffix, dbname='default'):
    '''
    Summary: Method to save decoder to DB -- saves to TE that original decoder came from
    Input param: task_entry_id: original task to save decoder to
    Input param: dec_obj: KF decoder new
    Input param: suffix:
    Output param: 
    '''

    te = dbfn.TaskEntry(task_entry_id)
    try:
        te_id = te.id
    except:
        dec_nm = te.name
        te_ix = re.search('te[0-9]',dec_nm)
        ix = te_ix.start() + 2
        sub_dec_nm = dec_nm[ix:]
        
        te_ix_end = sub_dec_nm.find('_')
        if te_ix_end == -1:
            te_ix_end = len(sub_dec_nm)
        te_id = int(sub_dec_nm[:te_ix_end])

    old_dec_obj = te.decoder_record
    if old_dec_obj is None:
        old_dec_obj = faux_decoder_obj(task_entry_id)
    trainbmi.save_new_decoder_from_existing(dec_obj, old_dec_obj, suffix=suffix, dbname=dbname)

def bin_(kin, neural_features, update_rate, desired_update_rate, only_neural=False):

    n = desired_update_rate/float(update_rate)
    if not only_neural:
        assert kin.shape[1] == neural_features.shape[1]
    
    ix_end = int(np.floor(neural_features.shape[1] / n)*n)

    if (n - round(n)) < 1e-5:
        n = int(n)
        if not only_neural:
            kin_ = kin[:, :ix_end].reshape(kin[:, :ix_end].shape[0], kin[:, :ix_end].shape[1]/n, n)
            bin_kf = np.mean(kin_, axis=2)

        nf_ = neural_features[:, :ix_end].reshape(neural_features[:, :ix_end].shape[0], neural_features[:, :ix_end].shape[1]/n, n)
        bin_nf = np.sum(nf_, axis=2)

        if only_neural:
            return bin_nf, desired_update_rate
        else:
            return bin_nf, bin_kf, desired_update_rate
    else:
        raise Exception('Desired rate %f not multiple of original rate %f', desired_update_rate, update_rate)

class faux_decoder_obj(object):
    def __init__(self, task_entry_id, *args,**kwargs):
        self.name = ''
        self.entry_id = task_entry_id



'''
Various graphical "environmental" or "world" add-ins to graphical tasks. 
'''
from .models import Group
from .xfm import Quaternion
from ..stereo_opengl.primitives import Sphere, Cylinder


class Box(Group):
    '''
    Construct a 3D wireframe box in the world to add some depth cue references
    '''
    def __init__(self, **kwargs):
        '''
        Constructor for Box 

        Parameters
        ----------
        kwargs: optional keyword arguments
            All passed to parent constructor

        Returns
        -------
        Box instance
        '''
        bcolor = (181/256., 116/256., 96/256., 1)
        sidelen = 10
        linerad=.1
        self.vert_box = Group([
            Cylinder(radius=linerad, height=sidelen, color=bcolor).translate(-sidelen/2, -sidelen/2, -sidelen/2),
            Cylinder(radius=linerad, height=sidelen, color=bcolor).translate(sidelen/2, -sidelen/2, -sidelen/2),
            Cylinder(radius=linerad, height=sidelen, color=bcolor).translate(sidelen/2, sidelen/2, -sidelen/2),
            Cylinder(radius=linerad, height=sidelen, color=bcolor).translate(-sidelen/2, sidelen/2, -sidelen/2)])
        self.hor_box = Group([
            Cylinder(radius=linerad, height=sidelen, color=bcolor).translate(-sidelen/2, -sidelen/2, -sidelen/2),
            Cylinder(radius=linerad, height=sidelen, color=bcolor).translate(sidelen/2, -sidelen/2, -sidelen/2),
            Cylinder(radius=linerad, height=sidelen, color=bcolor).translate(sidelen/2, sidelen/2, -sidelen/2),
            Cylinder(radius=linerad, height=sidelen, color=bcolor).translate(-sidelen/2, sidelen/2, -sidelen/2)])
        self.depth_box = Group([
            Cylinder(radius=linerad, height=sidelen, color=bcolor).translate(-sidelen/2, -sidelen/2, -sidelen/2),
            Cylinder(radius=linerad, height=sidelen, color=bcolor).translate(sidelen/2, -sidelen/2, -sidelen/2),
            Cylinder(radius=linerad, height=sidelen, color=bcolor).translate(sidelen/2, sidelen/2, -sidelen/2),
            Cylinder(radius=linerad, height=sidelen, color=bcolor).translate(-sidelen/2, sidelen/2, -sidelen/2)])
        self.hor_box.xfm.rotate = Quaternion.rotate_vecs((0,0,1), (1,0,0))
        self.depth_box.xfm.rotate = Quaternion.rotate_vecs((0,0,1), (0,1,0))
        self.box = Group([self.hor_box, self.depth_box, self.vert_box])
        super(Box, self).__init__([self.box], **kwargs)


#!/usr/bin/python
'''
Classes to determine the "goal" during a BMI task. Knowledge of the
task goal is required for many versions of assistive control (see assist.py) and
many versions of CLDA (see clda.py). 
'''
import numpy as np
from .. import mp_calc
import pickle

# TODO -- does ssm really need to be passed as an argument into __init__?
# maybe just make it an optional kwarg for the classes that really need it

class GoalCalculator(object):
    def reset(self):
        pass    

class Obs_Goal_Calc(GoalCalculator):
    def __init__(self, ssm=None, **kwargs):
        self.ssm = ssm
        import os
        self.pre_obs = True
        self.mid_speed = kwargs.pop('mid_targ_speed', 10)
        self.mid_targ_rad = kwargs.pop('mid_targ_rad', 6)
        self.targ_cnt = 0
        self.pre_obs_targ_state = None
        self.post_obs_targ_state = None

    def clear(self):
        self.pre_obs_targ_state = None
        self.post_obs_targ_state = None    
        print('CLEAR')    

    def __call__(self, target_pos, **kwargs):
        #Use q_start th
        pos = kwargs.pop('q_start')

        #if past obstacle midline: 
        
        if 'center_pos' in kwargs:
            obstacle_center = kwargs['center_pos'] + (target_pos - kwargs['center_pos'])*.5
            center = kwargs['center_pos']
        else:
            obstacle_center = target_pos/2.
            center = np.zeros((3, ))

        target_pos = target_pos.round(1)
        try:
            slope = -1*1./((target_pos[2] - center[2])/(target_pos[0]-center[0]))
        except:
            slope = np.inf
            
        #if ((np.abs(slope) != np.inf) and (np.abs(slope) != np.nan) and np.abs(slope)!=0):
        pre_obs = self.fcn_det(slope, obstacle_center, pos, center, target_pos)
            #print 'pre_obs: ', pre_obs, slope
        # else:
        #     #print 'division by zero!'
        #     if target_pos[0] ==0:
        #         #Division by zero
        #         pre_obs = False
        #         if np.abs(pos[2]) < (np.abs(obstacle_center[2])-.2): pre_obs = True
        #     elif target_pos[2] == 0:
        #         pre_obs = False
        #         if np.abs(pos[0]) < (np.abs(obstacle_center[0]) -.2): pre_obs = True
        #     else: 
        #         Exception('Not vertical or horiz. line causing divide by zero --> error')

        if pre_obs:
            if 1:
            #if self.pre_obs_targ_state is None:
                obs_ang = np.angle(obstacle_center[0]-center[0] + 1j*(obstacle_center[2]-center[2]))
                obs_r = np.abs((obstacle_center[0]-center[0]) + 1j*(obstacle_center[2]-center[2]))
                
                if self.ccw_fcn(pos, obstacle_center): 
                    targ_vect_ang = np.pi/2
                else:
                    targ_vect_ang = -1*np.pi/2

                target_state_pos = obstacle_center + self.mid_targ_rad*(np.array([np.cos(targ_vect_ang+obs_ang), 0, np.sin(targ_vect_ang+obs_ang)]))
                target_vel = self.mid_speed*np.array([np.cos(obs_ang), 0, np.sin(obs_ang)])
                target_state = np.hstack((target_state_pos, target_vel, 1)).reshape(-1, 1)
                self.pre_obs_targ_state = target_state
            else:
                target_state = self.pre_obs_targ_state

        else:
            if 1:
            #if self.post_obs_targ_state is None:
                target_vel = np.zeros_like(target_pos)
                offset_val = 1
                target_state = np.hstack([target_pos, target_vel, 1]).reshape(-1, 1)
                
                if self.pre_obs_targ_state is not None:
                    self.post_obs_targ_state = target_state
            else:
                target_state = self.post_obs_targ_state

        error = 0

        # if self.pre_obs != pre_obs:
        #     self.pre_obs = pre_obs
        #     print self.pre_obs, target_state

        return (target_state, error), True


    def fcn_det(self, slope, pt_on_line, test_pt, center, target):
        d_center = np.sqrt(np.sum((test_pt - center)**2))
        d_target = np.sqrt(np.sum((test_pt - target)**2))
        if d_center < d_target:
            return True
        else:
            return False


        # abs_pt = np.abs(pt_on_line)
        # abs_test = np.abs(test_pt)
        # slope = -1*np.abs(slope)

        # b = abs_pt[2] - slope*abs_pt[0]
       
        # if abs_test[2] +0.3 < (b + slope*abs_test[0]):
        #     return True
        # else:
        #     return False

        # zz = False
        # if 0 < b: zz = True

        # if test_pt[2] < 0:
        #     eps = -.2
        # else:
        #     eps = .2

        # test = False
        # if (test_pt[2]+eps) < ((slope*test_pt[0]) + b): test = True

        # if zz!=test:
        #     return True
        # else:
        #     return False

    def ccw_fcn(self, pos_test, pos_ref):
        theta1 = np.angle(pos_test[0] + 1j * pos_test[2])
        theta2 = np.angle(pos_ref[0]+ 1j*pos_ref[2])

        if pos_ref[0] < 0 and pos_test[0] < 0:
            if pos_ref[2] < 0 and pos_test[2] >0:
                theta2 += 2*np.pi
            elif pos_ref[2] > 0 and pos_test[2] < 0:
                theta1 += 2*np.pi
        return theta1 > theta2

class ZeroVelocityGoal(GoalCalculator):
    '''
    Assumes that the target state of the BMI is to move to the task-specified position with zero velocity
    '''
    def __init__(self, ssm=None):
        '''
        Constructor for ZeroVelocityGoal
    
        Parameters
        ----------
        ssm : state_space_models.StateSpace instance
            The state-space model of the Decoder that is being assisted/adapted. Not needed for this particular method
    
        Returns
        -------
        ZeroVelocityGoal instance
        '''
        try:
            self.ssm = ssm()
        except:
            self.ssm = ssm

    def __call__(self, target_pos, **kwargs):
        '''
        Calculate the goal state [p, 0, 1] where p is the n-dim position and 0 is the n-dim velocity
    
        Parameters
        ----------
        target_pos : np.ndarray
            Optimal position, in generalized coordinates (i.e., need not be cartesian coordinates)
        kwargs : optional kwargs
            These are ignored, just present for function call compatibility
    
        Returns
        -------
        np.ndarray
            (N, 1) indicating the target state
        '''

        # Add zero velocity if needed: 
        n_pos_vel_states = int(self.ssm.n_states) - 1
        if len(target_pos) < n_pos_vel_states :
            target_vel = np.zeros_like(target_pos)
            offset_val = 1
            target_state = np.hstack([target_pos, target_vel, 1]).reshape(-1, 1)
        elif len(target_pos) == n_pos_vel_states:
            target_state = np.hstack([target_pos, 1]).reshape(-1, 1)
        else:
            target_state = np.hstack(target_pos).reshape(-1, 1) # don't add offset
        error = 0
        return (target_state, error), True

class ZeroVelocityGoal_ismore(GoalCalculator):
    def __init__(self, ssm=None, pause_states=[]):
        try:
            self.ssm = ssm()
        except:
            self.ssm = ssm

        self.pause_states = pause_states

    def __call__(self, target_pos, state, **kwargs):
        if state in self.pause_states:
            target_state = kwargs['current_state']
            error = 0
        else:
            n_pos_vel_states = int(self.ssm.n_states) - 1
            if len(target_pos) < n_pos_vel_states :
                target_vel = np.zeros_like(target_pos)
                offset_val = 1
                target_state = np.hstack([target_pos, target_vel, 1]).reshape(-1, 1)
            else:
                target_state = np.hstack([target_pos, 1]).reshape(-1, 1)
            error = 0
        return (target_state, error), True            

class ZeroVelocityAccelGoal(ZeroVelocityGoal):
    '''
    Similar to ZeroVelocityGoal, but used for a second order system where you also want the goal acceleration to be zero.
    '''
    def __call__(self, target_pos, **kwargs):
        '''
        See ZeroVelocityGoal.__call__ for argument documentation
        '''
        target_vel = np.zeros_like(target_pos)
        target_acc = np.zeros_like(target_pos)
        offset_val = 1
        error = 0
        target_state = np.hstack([target_pos, target_vel, target_acc, 1])
        return (target_state, error), True        


class PlanarMultiLinkJointGoal(GoalCalculator, mp_calc.FuncProxy):
    '''
    Looks up goal configuration for a redundant system based on the endpoint goal and tries to find the closest solution.

    DEPRECATED: The method implemented has not been used for a long time, and is not the best method for achieving finding the "closest" config space solution as desired.
    '''
    def __init__(self, ssm, shoulder_anchor, kin_chain, multiproc=False, init_resp=None):
        def fn(target_pos, **kwargs):
            joint_pos = kin_chain.inverse_kinematics(target_pos, **kwargs)
            endpt_error = np.linalg.norm(kin_chain.endpoint_pos(joint_pos) - target_pos)

            target_state = np.hstack([joint_pos, np.zeros_like(joint_pos), 1])
            
            return target_state, endpt_error
        super(PlanarMultiLinkJointGoal, self).__init__(fn, multiproc=multiproc, waiting_resp='prev', init_resp=init_resp)

class PlanarMultiLinkJointGoalCached(GoalCalculator, mp_calc.FuncProxy):
    '''
    Determine the goal state of a redundant system by look-up-table, i.e. redundancy is collapsed 
    by arbitrary mapping between redudnant target space and configuration space

    TODO: since multiprocessing is not required for this class, it needs to do a better job of hiding the multiprocessing.
    '''
    def __init__(self, ssm, shoulder_anchor, kin_chain, multiproc=False, init_resp=None, **kwargs):
        '''
        Constructor for PlanarMultiLinkJointGoalCached

        Parameters
        ----------
        ssm : state_space_models.StateSpace instance
        shoulder_anchor : np.ndarray of shape (3,)
            Position of the manipulator anchor
        kin_chain : robot_arms.KinematicChain instance
            Object representing the kinematic chain linkages (D-H parameters)
        multiproc : bool, optional, default=False
            Should leave this false for this 'cached' method
        init_resp : None
            Ignore this if multiproc=False, as directed above.
        kwargs : optional keyword arguments
            Can pass in 'goal_cache_block' to specify from which task entry to grab the cache file. WARNING: this is currently commented out

        Returns
        -------
        PlanarMultiLinkJointGoalCached instance

        '''
        self.ssm = ssm
        self.shoulder_anchor = shoulder_anchor
        self.kin_chain = kin_chain
        if 0: #'goal_cache_block' in kwargs:
            goal_cache_block = kwargs.pop('goal_cache_block')
            self.cached_data = pickle.load(open('/storage/assist_params/tentacle_cache_%d.pkl' % goal_cache_block))
        else:
            self.cached_data = pickle.load(open('/storage/assist_params/tentacle_cache3.pkl'))

        def fn(target_pos, **kwargs):
            '''    Docstring    '''
            joint_pos = None
            for pos in self.cached_data:
                if np.linalg.norm(target_pos - np.array(pos)) < 0.001:
                    possible_joint_pos = self.cached_data[pos]
                    ind = np.random.randint(0, len(possible_joint_pos))
                    joint_pos = possible_joint_pos[ind]
                    break

            if joint_pos is None:
                raise ValueError("Unknown target position!: %s" % str(target_pos))

            target_state = np.hstack([joint_pos, np.zeros_like(joint_pos), 1])
            
            int_endpt_location = target_pos - shoulder_anchor
            endpt_error = np.linalg.norm(kin_chain.endpoint_pos(-joint_pos) - int_endpt_location)
            print(endpt_error)

            return (target_state, endpt_error)

        super(PlanarMultiLinkJointGoalCached, self).__init__(fn, multiproc=multiproc, waiting_resp='prev', init_resp=init_resp)

    def __call__(self, target_pos, **kwargs):
        '''
        Calculate the goal state [p, 0, 1] where p is the n-dim position and 0 is the n-dim velocity
        p is the configuration space position, which must be looked up based on the target_pos 
        (not a one-to-one mapping in general)
    
        Parameters
        ----------
        target_pos : np.ndarray
            Optimal position, in generalized coordinates (i.e., need not be cartesian coordinates)
        kwargs : optional kwargs
            These are ignored, just present for function call compatibility
    
        Returns
        -------
        np.ndarray
            (N, 1) indicating the target state
        '''
        joint_pos = None
        for pos in self.cached_data:
            if np.linalg.norm(target_pos - np.array(pos)) < 0.001:
                joint_pos = self.cached_data[pos]

        if joint_pos == None:
            raise ValueError("Unknown target position!: %s" % str(target_pos))

        target_state = np.hstack([joint_pos, np.zeros_like(joint_pos), 1])
        
        endpt_error = 0

        return (target_state, endpt_error), True



'''
NIDAQ Digital I/O interface code. Higher-level Python wrapper for pcidio 
'''

try:
    import pcidio
except:
    pass

class SendAll(object):
    '''
    Interface for sending all the task-generated data through the NIDAQ interface card
    '''
    def __init__(self, device="/dev/comedi0"):
        '''
        Constructor for SendAll

        Parameters
        ----------
        device : string, optional
            comedi device to open and use for data sending; only tested for NI PCI-6503

        Returns
        -------
        SendAll instance
        '''
        self.systems = dict()

        try:
            pcidio
        except:
            raise Exception('Cannot import pcidio. Did you run its build script?')
        if pcidio.init(device) != 0:
            raise ValueError("Could not initialize comedi system")
    
    def close(self):
        '''
        Release access to the NIDAQ card so that other processes can access the device later.
        '''
        if pcidio.closeall() != 0:
            raise ValueError("Unable to close comedi system")

    def register(self, system, dtype):
        '''
        Send information about the registration system (name and datatype) in string form, one byte at a time.

        Parameters
        ----------
        system : string
            Name of the system being registered
        dtype : np.dtype instance
            Datatype of incoming data, for later decoding of the binary data during analysis

        Returns
        -------
        None
        '''
        print("nidaq register %s" % system)

        # Save the index of the system being registered (arbitrary number corresponding to the order in which systems were registered)
        self.systems[system] = pcidio.register_sys(system, str(dtype.descr))

    def sendMsg(self, msg):
        '''
        Send a string mesasge to the recording system, e.g., as related to the task_msgs HDF table

        Parameters
        ----------
        msg : string
            Message to send

        Returns
        -------
        None
        '''
        pcidio.sendMsg(str(msg))

    def sendRow(self, system, idx):
        '''
        The function pcidio.sendRow tries to send an array through the card one byte at a time. 
        It's unclear how the underlying code is supposed to get the 'idx' pointer that is required to use this function...
        .. and this function appears to be unused and probably should be removed. 
        Abandoned in place in case there is a future unknown use for this functionality
        '''
        if system in self.systems:
            pcidio.sendRow(self.systems[system], idx)

    def rstart(self, state):
        '''
        Remotely start recording from the plexon system

        Parameters
        ----------
        state : int
            0 or 1 depending on whether you want the system to start or stop
            For the plexon system, this is actually not used, and instead the comedi python bindings generate the required pulse.

        Returns
        -------
        None
        '''
        print("Sending rstart command")
        pcidio.rstart(state)

    def send(self, system, data):
        '''
        Send data through the DIO device

        Parameters
        ----------
        system : string
            Name of system where the data originated
        data : object
            Data to send. Must have a '.tostring()' attribute

        Returns
        -------
        None
        '''
        if system in self.systems:
            pcidio.sendData(self.systems[system], data.tostring())

class SendRow(SendAll):
    '''
    Send the number of rows to the plexon system. Used by riglib.hdfwriter.PlexRelayWriter, which is never actually used anywhere....
    '''
    def send(self, system, data):
        '''
        Send data to a registered system

        Parameters
        ----------
        system : string
            Name of system where the data originated
        data : object
            Argument is ignored, since only the count is sent and not the actual data

        Returns
        -------
        None
        '''
        if system in self.systems:
            pcidio.sendRowCount(self.systems[system])

class SendRowByte(SendAll):
    '''
    Send only an 8-bit data word corresponding to the 8 lower bits of the current row number of the HDF table
    '''
    def send(self, system, data):
        '''
        Send data to a registered system

        Parameters
        ----------
        system : string
            Name of system where the data originated
        data : object
            Argument is ignored, since only the count is sent and not the actual data

        Returns
        -------
        None
        '''
        if system in self.systems:
            pcidio.sendRowByte(self.systems[system])



# -*- coding: utf-8 -*-
"""
Random functions that may be useful elsewhere (or necessary)
current version: 1.2.0 --- 08/04/2016

@author: Mitch Frankel - Blackrock Microsystems
"""

"""
Version History:
v1.0.0 - 07/05/2016 - initial release
v1.1.0 - 07/12/2016 - minor editing changes to print statements and addition of version control
v1.2.0 - 08/04/2016 - minor modifications to allow use of Python 2.6+
"""

from os             import getcwd, path

# Version control
brmiscfxns_ver = "1.2.0"

# Patch for use with Python 2.6+
try: input = raw_input
except NameError: pass

def openfilecheck(open_mode, file_name='', file_ext='', file_type=''):
    """
    :param open_mode: {str} method to open the file (e.g., 'rb' for binary read only)
    :param file_name: [optional] {str} full path of file to open
    :param file_ext:  [optional] {str} file extension (e.g., '.nev')
    :param file_type: [optional] {str} file type for use when browsing for file (e.g., 'Blackrock NEV Files')
    :return: {file} opened file
    """

    while True:
        if not file_name:  # no file name passed
            from qtpy.QtWidgets import QFileDialog, QApplication
            # Ask user to specify a file path or browse
            file_name = eval(input("Enter complete " + file_ext + " file path or hit enter to browse: "))

            if not file_name:
                if 'app' not in locals(): app = QApplication([])
                if not file_ext: file_type = 'All Files'
                file_name = QFileDialog.getOpenFileName(QFileDialog(), "Select File", getcwd(),
                                                        file_type + " (*" + file_ext + ")")

        # Ensure file exists (really needed for users type entering)
        if path.isfile(file_name):
            # Ensure given file matches file_ext
            if file_ext:
                _, fext = path.splitext(file_name)

                # check for * in extension
                if file_ext[-1] == '*': test_extension = file_ext[:-1]
                else:                   test_extension = file_ext

                if fext[0:len(test_extension)] != test_extension:
                    file_name = ''
                    print(("\n*** File given is not a " + file_ext + " file, try again ***\n"))
                    continue
            break
        else:
            file_name = ''
            print("\n*** File given does exist, try again ***\n")

    print(('\n' + file_name.split('/')[-1] + ' opened'))
    return open(file_name, open_mode)


def checkequal(iterator):
    try:
        iterator = iter(iterator)
        first = next(iterator)
        return all(first == rest for rest in iterator)
    except StopIteration:
        return True



# Generated by Django 3.1.4 on 2023-11-08 15:03

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('tracker', '0011_auto_20211017_1548'),
    ]

    operations = [
        migrations.AddField(
            model_name='taskentry',
            name='rig_name',
            field=models.TextField(default='pagaiisland2'),
        ),
    ]



'''
Interact with a peripheral 'button' device hooked up to an FTDI chip
'''

import threading
import queue
import ftdi
import time

class Button(threading.Thread):
    ''' Docstring '''
    def __init__(self):
        '''
        Docstring

        Parameters
        ----------

        Returns
        -------
        '''
        super(Button, self).__init__()
        self.port = None
        port = ftdi.ftdi_new()
        usb_open = ftdi.ftdi_usb_open_string(port, "s:0x403:0x6001:2eb80091")
        assert usb_open == 0, ftdi.ftdi_get_error_string(port)
        
        ftdi.ftdi_set_bitmode(port, 0xFF, ftdi.BITMODE_BITBANG)
        self.port = port
        self.queue = queue.Queue()
        self.daemon = True
        self.start()
       
    def run(self):
        '''
        Docstring

        Parameters
        ----------

        Returns
        -------
        '''
        last = None
        while True:
            k = self._check()
            if last == 0 and k != 0:
                self.queue.put(k)
            last = k
            time.sleep(0.01)
    
    def _check(self):
        '''
        Docstring

        Parameters
        ----------

        Returns
        -------
        '''
        test = ' '
        ftdi.ftdi_read_pins(self.port, test)
        return ord(test)
    
    def pressed(self):
        '''
        Docstring

        Parameters
        ----------

        Returns
        -------
        '''
        try:
            return self.queue.get_nowait()
        except:
            return None
    
    def __del__(self):
        '''
        Docstring

        Parameters
        ----------

        Returns
        -------
        '''
        if self.port is not None:
            ftdi.ftdi_disable_bitbang(self.port)
            ftdi.ftdi_usb_close(self.port)
            ftdi.ftdi_deinit(self.port)

if __name__ == "__main__":
    import time
    btn = Button()
    while True:
        k = btn.pressed()
        if k is not None:
            print(k)



'''
Various types of "assist", i.e. different methods for shared control
between neural control and machine control. Only applies in cases where
some knowledge of the task goals is available. 
'''

import numpy as np 
from ..bmi import feedback_controllers
from utils.angle_utils import *
from utils.constants import *

class Assister(object):
    '''
    Parent class for various methods of assistive BMI. Children of this class 
    can compute an "optimal" input to the system, which is mixed in with the input
    derived from the subject's neural input. The parent exists primarily for 
    interface standardization and type-checking.
    '''
    def calc_assisted_BMI_state(self, current_state, target_state, assist_level, mode=None, **kwargs):
        '''
        Main assist calculation function

        Parameters
        ----------
        current_state: np.ndarray of shape (n_states, 1)
            Vector representing the current state of the prosthesis 
        target_state: np.ndarray of shape (n_states, 1)
            Vector representing the target state of the prosthesis, i.e. the optimal state for the prosthesis to be in
        assist_level: float
            Number indicating the level of the assist. This can in general have arbitrary units but most assisters
            will have this be a number in the range (0, 1) where 0 is no assist and 1 is full assist
        mode: hashable type, optional, default=None
            Indicator of which mode of the assistive controller to use. When applied, this 'mode' is used as a dictionary key and must be hashable
        kwargs: additional keyword arguments
            These are ignored

        Returns
        -------
        '''
        pass

    def __call__(self, *args, **kwargs):
        '''
        Wrapper for self.calc_assisted_BMI_state
        '''
        return self.calc_assisted_BMI_state(*args, **kwargs)

class FeedbackControllerAssist(Assister):
    '''
    Assister where the machine control is an LQR controller, possibly with different 'modes' depending on the state of the task
    '''
    def __init__(self, fb_ctrl, style='additive'):
        '''
        Parameters
        ----------
        fb_ctrl : feedback_controllers.FeedbackController instance
            TODO

        Returns
        -------
        FeedbackControllerAssist instance
        '''
        self.fb_ctrl = fb_ctrl
        self.style = style
        assert self.style in ['additive', 'mixing', 'additive_cov']

    def calc_assisted_BMI_state(self, current_state, target_state, assist_level, mode=None, **kwargs):
        '''
        See docs for Assister.calc_assisted_BMI_state
        '''
        if self.style == 'additive':
            Bu = assist_level * self.fb_ctrl(current_state, target_state, mode=mode)
            return dict(Bu=Bu, assist_level=0)
        elif self.style == 'mixing':
            x_assist = self.fb_ctrl.calc_next_state(current_state, target_state, mode=mode)
            return dict(x_assist=x_assist, assist_level=assist_level)
        elif self.style == 'additive_cov':
            F = self.get_F(assist_level)
            return dict(F=F, x_target=target_state)            

class FeedbackControllerAssist_StateSpecAssistLevels(FeedbackControllerAssist):
    '''
    Assister where machine controller is LQR controller, but different assist_levels for 
    different control variables (e.g. X,Y,PSI in ArmAssist vs. Rehand)
    '''
    def __init__(self, fb_ctrl, style='additive', **kwargs):
        super(FeedbackControllerAssist_StateSpecAssistLevels, self).__init__(fb_ctrl, style)
        
        # Currently this assister assumes that plant is IsMore Plant: 
        self.assist_level_state_ix = dict()
        self.assist_level_state_ix[0] = np.array([0, 1, 2, 7, 8, 9]) # ARM ASSIST
        self.assist_level_state_ix[1] = np.array([3, 4, 5, 6, 10, 11, 12, 13]) # REHAND
        

    def calc_assisted_BMI_state(self, current_state, target_state, assist_level, mode=None, **kwargs):
        if self.style == 'additive':
            Bu = self.fb_ctrl(current_state, target_state, mode=mode)
            for ia, al in enumerate(assist_level):
                Bu[self.assist_level_state_ix[ia]] = al*Bu[self.assist_level_state_ix[ia]]
            return dict(Bu=Bu, assist_level=0)
        
        elif self.style == 'mixing':
            x_assist = self.fb_ctrl.calc_next_state(current_state, target_state, mode=mode)
            return dict(x_assist=x_assist, assist_level=assist_level, assist_level_ix=self.assist_level_state_ix)
        

class SSMLFCAssister(FeedbackControllerAssist):
    '''
    An LFC assister where the state-space matrices (A, B) are specified from the Decoder's 'ssm' attribute
    '''
    def __init__(self, ssm, Q, R, **kwargs):
        '''
        Constructor for SSMLFCAssister

        Parameters
        ----------
        ssm: riglib.bmi.state_space_models.StateSpace instance
            The state-space model's A and B matrices represent the system to be controlled
        args: positional arguments
            These are ignored (none are necessary)
        kwargs: keyword arguments
            The constructor must be supplied with the 'kin_chain' kwarg, which must have the attribute 'link_lengths'
            This is specific to 'KinematicChain' plants.

        Returns
        -------
        SSMLFCAssister instance

        '''        
        if ssm is None:
            raise ValueError("SSMLFCAssister requires a state space model!")

        A, B, W = ssm.get_ssm_matrices()
        self.lqr_controller = feedback_controllers.LQRController(A, B, Q, R)



import numpy as np
from riglib import bmi
from riglib.bmi import state_space_models, train, extractor

bmi_update_rates = [10, 20, 30, 60, 120, 180]

bmi_algorithms = dict(
    KFDecoder=bmi.train.train_KFDecoder,
    PPFDecoder=bmi.train.train_PPFDecoder,
    OneDimLFPDecoder=bmi.train.create_onedimLFP,
    LinearDecoder=bmi.train.create_lindecoder,
)

bmi_training_pos_vars = [
    'cursor',
    'joint_angles',
    'plant_pos',  # used for ibmi tasks
    'mouse_state',
    'decoder_state',
]

## State-space models for BMIs
joint_2D_state_space = bmi.state_space_models.StateSpaceNLinkPlanarChain(n_links=2, w=0.01)
tentacle_2D_state_space = bmi.state_space_models.StateSpaceNLinkPlanarChain(n_links=4, w=0.01)
endpt_2D_state_space = bmi.state_space_models.StateSpaceEndptVel2D()

# ## Velocity SSMs
# from riglib.bmi.state_space_models import offset_state, State
# endpt_2D_states = [State('hand_px', stochastic=False, drives_obs=False, min_val=-25., max_val=25., order=0),
#                    State('hand_py', stochastic=False, drives_obs=False, order=0),
#                    State('hand_pz', stochastic=False, drives_obs=False, min_val=-14., max_val=14., order=0),
#                    State('hand_vx', stochastic=True,  drives_obs=True, order=1),
#                    State('hand_vy', stochastic=False, drives_obs=False, order=1),
#                    State('hand_vz', stochastic=True,  drives_obs=True, order=1),
#                    offset_state]
# endpt_2D_state_space = state_space_models.LinearVelocityStateSpace(endpt_2D_states)

bmi_state_space_models=dict(
    Endpt2D=endpt_2D_state_space,
    Tentacle=tentacle_2D_state_space,
    Joint2L=joint_2D_state_space,
)

extractors = dict(
    spikecounts = extractor.BinnedSpikeCountsExtractor,
    LFPpowerMTM = extractor.LFPMTMPowerExtractor,
    direct = extractor.DirectObsExtractor,
)

kin_extractors = dict(
    pos_vel=train.get_plant_pos_vel,
    null=train.null_kin_extractor,
)

default_extractor = "spikecounts"




'''
Generic class for implementing filters describable by rational z-transforms (ratio of polynomials)
'''
from scipy.signal import sigtools, lfilter
import numpy as np

class Filter(object):
    def __init__(self, b=[], a=[1.]):
        '''
        Constructor for Filter

        Parameters
        ----------
        b : array_like
            The numerator coefficient vector in a 1-D sequence.
        a : array_like
            The denominator coefficient vector in a 1-D sequence.  If ``a[0]``
            is not 1, then both `a` and `b` are normalized by ``a[0]``.

        Returns
        -------
        Filter instance
        '''
        self.b = np.array(b)
        self.a = np.array(a)

        # normalize the constants
        self.b /= self.a[0]
        self.a /= self.a[0]

        self.zi = np.zeros(max(len(a), len(b))-1)


    def __call__(self, samples):
        '''
        Run the filter parameters on the most recent set of samples

        Parameters
        ----------
        samples : np.ndarray of shape (N,)
            samples to filter (x input)

        Returns
        -------
        np.ndarray
            Most recent N outputs of the filter
        '''
        # promote scalars to arrays so the lfilter function doesn't complain
        if isinstance(samples, (float, int)):
            samples = np.array([samples])

        filt_output, self.zi = lfilter(self.b, self.a, samples, zi=self.zi)
        return filt_output


from riglib import experiment
from built_in_tasks.manualcontrolmultitasks import ManualControl
from built_in_tasks.othertasks import LaserConditions
from riglib.stereo_opengl.window import WindowDispl2D
from features.peripheral_device_features import KeyboardControl, MouseControl
import features.sync_features as sync_features
from features.laser_features import CrystaLaser
from features.video_recording_features import E3Video
from riglib.e3vision import E3VisionInterface
import numpy as np
import time

import unittest


def init_exp(base_class, feats, seq=None, **kwargs):
    blocks = 2
    trials = 2
    trial_length = 5
    frequencies = np.array([.5])
    Exp = experiment.make(base_class, feats=feats)
    if seq is not None:
        exp = Exp(seq, **kwargs)
    else:
        exp = Exp(**kwargs)
    exp.init()
    return exp

class TestKeyboardControl(unittest.TestCase):

    @unittest.skip("msg")
    def test_exp(self):
        exp = init_exp(ManualControl, [KeyboardControl, WindowDispl2D])
        exp.run()

class TestMouseControl(unittest.TestCase):

    @unittest.skip("msg")
    def test_exp(self):
        exp = init_exp(ManualControl, [MouseControl, WindowDispl2D])
        exp.run()

class TestLaser(unittest.TestCase):
    
    def test_digital_wave(self):
        from riglib.gpio import TestGPIO, DigitalWave
        gpio = TestGPIO()
        laser1 = DigitalWave(gpio, mask=2)
        laser1.set_pulse(1, 0)
        self.assertCountEqual(laser1.edges, [0, 1])
        laser1.set_pulse(1, 1)
        self.assertCountEqual(laser1.edges, [0, 1])
        laser1.set_square_wave(1, 5)
        self.assertCountEqual(laser1.edges, np.linspace(0, 5.0, 11))
        laser1.run() # run on the same thread so we can make changes to the test gpio state
        print(gpio.value[1,:])
        self.assertEqual(sum(gpio.value[1,:]), 6) # 1 Hz over 5 seconds has 6 positive edges including the last one at 5 s

        # Test duty cycle
        laser2 = DigitalWave(gpio, mask=2)
        laser2.set_square_wave(1, 5, duty_cycle=0.2)
        expected = [0. , 0.2, 1. , 1.2, 2. , 2.2, 3. , 3.2, 4. , 4.2, 5. ]
        self.assertCountEqual(laser2.edges, expected)

        # Test phase delay
        laser2 = DigitalWave(gpio, mask=2)
        laser2.set_square_wave(1, 5, phase_delay=0.2)
        self.assertCountEqual(laser2.edges, np.linspace(0.2, 5.2, 11))

    def test_convert_masked_data_to_pins(self):
        from riglib.gpio import convert_masked_data_to_pins
        pins, values = convert_masked_data_to_pins(1<<12, 1<<12)
        self.assertEqual(pins, [12])
        self.assertTrue(values)

    @unittest.skip("Need arduino connected for this to pass")
    def test_arduino(self):
        from riglib.gpio import ArduinoGPIO, DigitalWave
        gpio = ArduinoGPIO('/dev/crystalaser')
        ch = 12
        laser = DigitalWave(gpio, mask=1<<ch)
        laser.set_square_wave(5, 10)
        laser.start()
        laser.join()
        laser = DigitalWave(gpio, mask=1<<ch)
        laser.set_edges([0], False)
        laser.start()
        laser.join()

    @unittest.skip("Need arduino connected for this to pass")
    def test_laser_conditions(self):
        Exp = experiment.make(LaserConditions, [CrystaLaser])
        exp = Exp(LaserConditions.single_laser_pulse(nreps=2))
        exp.init()
        exp.run()

    def test_dio_pulse_width(self):
        from riglib.gpio import ArduinoGPIO, DigitalWave
        import time
        gpio = ArduinoGPIO('/dev/crystalaser')

        pulse_widths = [0.00001, 0.00005, 0.0001, 0.0002, 0.0005, 0.001]
        reps = 20
        durations = np.zeros((len(pulse_widths),reps))
        for width_idx, width in enumerate(pulse_widths):
            for n in range(reps):
                t0 = time.perf_counter()
                gpio.write_many(1<<12, 1<<12)
                while (time.perf_counter() - t0 < width):
                    pass
                durations[width_idx, n] = time.perf_counter() - t0
                gpio.write_many(1<<12, 0)
                time.sleep(0.1)

        print(np.mean(durations, axis=1))
        print(np.std(durations, axis=1))

        time.sleep(1)

        seq = LaserConditions.single_laser_pulse(nreps=reps, duration=pulse_widths, uniformsampling=False, ascending=True)
        for idx, powers, edges in seq:
            edges = edges[0]
            wave = DigitalWave(gpio, mask=1<<12)
            wave.set_edges(edges, True)
            wave.start()
            wave.join()
            time.sleep(0.1)

class TestSync(unittest.TestCase):

    def test_dictionary(self):
        default_dict = sync_features.rig1_sync_params['event_sync_dict']
        self.assertEqual(default_dict['TARGET_ON'] + 4, sync_features.encode_event(default_dict, 'TARGET_ON', 4))
        for k in default_dict.keys():
            event_data = 0
            encode = sync_features.encode_event(default_dict, k, event_data)
            decode = sync_features.decode_event(default_dict, encode)
            self.assertEqual(decode[0], k)    
            self.assertEqual(decode[1], event_data)

class TestE3Video(unittest.TestCase):

    # def test_interface(self):
    #     e3v = E3VisionInterface()
    #     e3v.update_camera_status()
    #     e3v.start_rec()
    #     time.sleep(5)
    #     e3v.stop_rec()

    def test_feature(self):
        exp = init_exp(experiment.Experiment, [E3Video], saveid=0)
        exp.run()
        time.sleep(1)
        exp.state = None

if __name__ == '__main__':
    unittest.main()


'''
Features which have task-like functionality w.r.t. task...
'''

import random
import numpy as np
from scipy.spatial.transform import Rotation as R
from riglib.experiment import traits
from built_in_tasks.target_graphics import *

class Autostart(traits.HasTraits):
    '''
    Automatically begins the trial from the wait state, 
    with a random interval drawn from `rand_start`. Doesn't really
    work if there are multiple trials in between wait states.
    '''
    rand_start = traits.Tuple((0., 0.), desc="Start interval")
    exclude_parent_traits = ['wait_time']

    def _start_wait(self):
        '''
        At the start of the 'wait' state, determine how long to wait before starting the trial
        by drawing a sample from the rand_start interval
        '''
        s, e = self.rand_start
        self.wait_time = random.random()*(e-s) + s
        super(Autostart, self)._start_wait()
        
    def _test_start_trial(self, ts):
        '''
        Test if the required random wait time has passed
        '''
        return ts > self.wait_time and not self.pause

class AdaptiveGenerator(object):
    '''
    Deprecated--this class appears to be unused
    '''
    def __init__(self, *args, **kwargs):
        super(AdaptiveGenerator, self).__init__(*args, **kwargs)
        assert hasattr(self.gen, "correct"), "Must use adaptive generator!"

    def _start_reward(self):
        self.gen.correct()
        super(AdaptiveGenerator, self)._start_reward()
    
    def _start_incorrect(self):
        self.gen.incorrect()
        super(AdaptiveGenerator, self)._start_incorrect()


class IgnoreCorrectness(object):
    '''Deprecated--this class appears to be unused and not compatible with Sequences
    Allows any response to be correct, not just the one defined. Overrides for trialtypes'''
    def __init__(self, *args, **kwargs):
        super(IgnoreCorrectness, self).__init__(*args, **kwargs)
        if hasattr(self, "trial_types"):
            for ttype in self.trial_types:
                del self.status[ttype]["%s_correct"%ttype]
                del self.status[ttype]["%s_incorrect"%ttype]
                self.status[ttype]["correct"] = "reward"
                self.status[ttype]["incorrect"] = "penalty"

    def _test_correct(self, ts):
        return self.event is not None

    def _test_incorrect(self, ts):
        return False


class MultiHoldTime(traits.HasTraits):
    '''
    Deprecated--Use RandomDelay instead. 
    Allows the hold time parameter to be multiple values per target in a given sequence chain. For instance,
    center targets and peripheral targets can have different hold times.
    '''

    multi_hold_time = traits.List([.2,], desc="Length of hold required at targets before next target appears. \
        Can be a single number or a list of numbers to apply to each target in the sequence (center, out, etc.)")
    exclude_parent_traits = ['hold_time']
    
    def _test_hold_complete(self, time_in_state):
        '''
        Test whether the target is held long enough to declare the
        trial a success

        Possible options
            - Target held for the minimum requred time (implemented here)
            - Sensorized object moved by a certain amount
            - Sensorized object moved to the required location
            - Manually triggered by experimenter
        '''
        if len(self.multi_hold_time) == 1:
            multi_hold_time = self.multi_hold_time[0]
        else:
            multi_hold_time = self.multi_hold_time[self.target_index]
        return time_in_state > multi_hold_time

class Progressbar_fixation(traits.HasTraits):

    progress_bar_color = traits.OptionsList("blue", *target_colors, desc="Color of the eye target", bmi3d_input_options=list(target_colors.keys()))

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Instantiate the targets
        instantiate_targets = kwargs.pop('instantiate_targets', True)
        if instantiate_targets:

            # Target 1 and 2 are for saccade. Target 3 is for hand
            self.bar = VirtualCircularTarget(target_radius=0.5, target_color=target_colors[self.progress_bar_color], starting_pos=[-5,0,-5])

    def _start_wait(self):
        super()._start_wait()

        if self.calc_trial_num() == 0:

            for model in self.bar.graphics_models:
                self.add_model(model)
                self.bar.hide()

    def _start_fixation(self):
        super()._start_fixation()
        self.fixation_frame_index = 0

    def _while_fixation(self):
        super()._while_fixation()
        self.fixation_frame_index += 1
        fixation_time = self.fixation_frame_index*2/self.fps
        if fixation_time <= self.hold_time:
            radius = fixation_time/self.hold_time*(self.fixation_radius-self.fixation_radius_buffer)
        else:
            radius = (self.fixation_radius-self.fixation_radius_buffer)

        if hasattr(self, 'bar'):
            for model in self.bar.graphics_models:
                self.remove_model(model)
            del self.bar

        self.bar = VirtualCircularTarget(target_radius=radius, target_color=target_colors[self.progress_bar_color], \
                                         starting_pos=self.targs[self.target_index]+[0,-0.5,0]) # [0,-0.5,0] is for visualization
        for model in self.bar.graphics_models:
            self.add_model(model)
        self.bar.show()

    def _start_targ_transition(self):
        super()._start_targ_transition()
        self.bar.hide()


class RandomDelay(traits.HasTraits):
    '''
    Replaces 'delay_time' with 'rand_delay', an interval on which the delay period is selected uniformly.
    '''
    
    rand_delay = traits.Tuple((0., 0.), desc="Delay interval")
    exclude_parent_traits = ['delay_time']
    prob_catch_trials = traits.Float(0., desc="Probability of catch trials")
    short_delay_catch_trials = traits.List([.2,], desc="Delay intervals for catch trials")

    def _start_wait(self):
        '''
        At the start of the 'wait' state, draw a sample from the rand_delay interval for this trial.
        '''

        # Catch trial condition
        if random.random() < self.prob_catch_trials:
            self.delay_time = random.choice(self.short_delay_catch_trials)

        # Normal trial condition
        else:
            s, e = self.rand_delay
            self.delay_time = random.random()*(e-s) + s
        super()._start_wait()

class TransparentDelayTarget(traits.HasTraits):
    '''
    Feature to make the delay period show a semi-transparent target rather than the full target. Used 
    for training the go cue. Gradually increase the alpha from 0 to 0.75 once a long enough delay 
    period has been established.
    '''

    delay_target_alpha = traits.Float(0.25, desc="Transparency of the next target during delay periods")

    def _start_delay(self):
        super()._start_delay()

        # Set the alpha of the next target
        next_idx = (self.target_index + 1)
        if next_idx < self.chain_length:
            target = self.targets[next_idx % 2]
            self._old_target_color = np.copy(target.sphere.color)
            new_target_color = list(target.sphere.color)
            new_target_color[3] = self.delay_target_alpha
            target.sphere.color = tuple(new_target_color)

    def _start_target(self):
        super()._start_target()

        # Reset the transparency of the current target
        if self.target_index > 0:
            target = self.targets[self.target_index % 2]
            target.sphere.color = self._old_target_color

class PoissonWait(traits.HasTraits):
    '''
    Draw each trial's wait time from a poisson random distribution    
    '''
    
    poisson_mu = traits.Float(0.5, desc="Mean duration between trials (s)")
    exclude_parent_traits = ['wait_time']

    def _parse_next_trial(self):
        self.wait_time = np.random.exponential(self.poisson_mu)
        super()._parse_next_trial()

class IncrementalRotation(traits.HasTraits):
    '''
    Gradually change the perturbation rotation over trials
    '''
    exclude_parent_traits = ['pertubation_rotation', 'perturbation_rotation_z', 'perturbation_rotation_x']

    init_rotation_y  = traits.Float(0.0, desc="initial rotation about bmi3d y-axis in degrees")
    final_rotation_y = traits.Float(0.0, desc="final rotation about bmi3d y-axis in degrees")

    init_rotation_z  = traits.Float(0.0, desc="initial rotation about bmi3d z-axis in degrees")
    final_rotation_z = traits.Float(0.0, desc="final rotation about bmi3d z-axis in degrees")
    
    init_rotation_x  = traits.Float(0.0, desc="inital rotation about bmi3d x-axis in degrees")
    final_rotation_x = traits.Float(0.0, desc="final rotation about bmi3d x-axis in degrees")

    delta_rotation_y = traits.Float(0.0, desc="rotation step size about bmi3d y-axis in degrees")
    delta_rotation_z = traits.Float(0.0, desc="rotation step size about bmi3d z-axis in degrees")
    delta_rotation_x = traits.Float(0.0, desc="rotation step size about bmi3d x-axis in degrees")

    trials_per_increment = traits.Int(1, desc="number of successful trials per rotation step")
    final_tracking_out_time = traits.Float(1.6, desc="Time allowed to be tracking outside the target for final rotation") # AKA tolerance time

    def init(self):    
        super().init()
        self.num_trials_success = 0
        self.num_increments_y = 0
        self.num_increments_z = 0
        self.num_increments_x = 0

        if self.final_rotation_y != self.init_rotation_y:
            self.num_increments_y = int( (self.final_rotation_y-self.init_rotation_y) / self.delta_rotation_y+1 )
        if self.final_rotation_z != self.init_rotation_z:
            self.num_increments_z = int( (self.final_rotation_z-self.init_rotation_z) / self.delta_rotation_z+1 )
        if self.final_rotation_x != self.init_rotation_x:
            self.num_increments_x = int( (self.final_rotation_x-self.init_rotation_x) / self.delta_rotation_x+1 )

        self.max_num_increments = np.max([self.num_increments_y, self.num_increments_z, self.num_increments_x])
        self.pertubation_rotation = self.init_rotation_y
        self.perturbation_rotation_z = self.init_rotation_z
        self.perturbation_rotation_x = self.init_rotation_x

        print("Y", self.pertubation_rotation, "Z", self.perturbation_rotation_z, "X", self.perturbation_rotation_x)
    
    def incremental_start_wait(self):
        # determine the current rotation step
        num_deltas = int(self.num_trials_success / self.trials_per_increment)

        # increment the current perturbation rotation by delta
        self.pertubation_rotation = self.init_rotation_y + self.delta_rotation_y*num_deltas
        self.perturbation_rotation_z = self.init_rotation_z + self.delta_rotation_z*num_deltas
        self.perturbation_rotation_x = self.init_rotation_x + self.delta_rotation_x*num_deltas

        # change tracking out time of final rotation
        if num_deltas+1 == self.max_num_increments:
            self.tracking_out_time = self.final_tracking_out_time

        # stop incrementing once final perturbation rotation reached
        if self.num_trials_success >= self.num_increments_y * self.trials_per_increment:
            self.pertubation_rotation = self.final_rotation_y
        if self.num_trials_success >= self.num_increments_z * self.trials_per_increment:
            self.perturbation_rotation_z = self.final_rotation_z
        if self.num_trials_success >= self.num_increments_x * self.trials_per_increment:
            self.perturbation_rotation_x = self.final_rotation_x
        
        print("Y", self.pertubation_rotation, "Z", self.perturbation_rotation_z, "X", self.perturbation_rotation_x)
        print(self.tracking_out_time, "tracking out")
    
    def _start_wait(self):
        super()._start_wait()
        self.incremental_start_wait()

    def _start_wait_retry(self):
        super()._start_wait_retry()
        self.incremental_start_wait()

    def _start_reward(self):
        super()._start_reward()
        self.num_trials_success += 1


'''
Features for interacting with Blackrock's Neuroport neural recording system
'''

import time
import fnmatch
from .neural_sys_features import CorticalData, CorticalBMI
from riglib import blackrock

class RelayBlackrock(object):
    '''Sends full data directly into the Blackrock system.'''

    def init(self):
        '''
        Secondary init function. See riglib.experiment.Experiment.init()
        Prior to starting the task, this 'init' creates a sink for the NIDAQ card
        '''

        from riglib import sink
        sink_manager = sink.SinkManager.get_instance()
        self.nidaq = sink_manager.start(self.ni_out)
        super(RelayBlackrock, self).init()

    @property
    def ni_out(self):
        '''
        Docstring

        Parameters
        ----------

        Returns
        -------
        '''
        from riglib.dio import nidaq
        print('nidaq.SendAll', nidaq.SendAll)
        return nidaq.SendAll

    @property    
    def blackrockfiles(self):
        '''Finds the blackrock files --- .nev and .nsx (.ns1 through .ns6) 
        --- that are most likely associated with the current task based on 
        the time at which the task ended and the "last modified" time of the 
        files located at /storage/blackrock/.
        '''
        import os, sys, glob, time
        if len(self.event_log) < 1:
            return None

        start = self.event_log[-1][2]
        
        files = []
        for ext in ['.nev', '.ns1', '.ns2', '.ns3', '.ns4', '.ns5', '.ns6']:
            pattern = "/storage/blackrock/*" + ext

            #matches = sorted(glob.glob(pattern), key=lambda f: abs(os.stat(f).st_mtime - start))
            matches = []
            for root, dirnames, filenames in os.walk('/storage/blackrock'):
                for filename in fnmatch.filter(filenames, '*' + ext):
                    matches.append(os.path.join(root, filename))

            sorted_matches = sorted(matches, key=lambda f: abs(os.stat(f).st_mtime - start))

            if len(sorted_matches) > 0:
                tdiff = os.stat(sorted_matches[0]).st_mtime - start
                if abs(tdiff) < 60:
                     files.append(sorted_matches[0])

        return files
    
    def run(self):
        '''
        Code to execute immediately prior to the beginning of the task FSM executing, or after the FSM has finished running. 
        See riglib.experiment.Experiment.run(). This 'run' method stops the NIDAQ sink after the FSM has finished running
        '''

        try:
            super(RelayBlackrock, self).run()
        finally:
            self.nidaq.stop()

            # Remotely stop the recording on the blackrock box
            import comedi
            import config
            import time
            com = comedi.comedi_open("/dev/comedi0")
            time.sleep(0.5)
            # strobe pin should already be low

            # set last data pin ("D15"; 16th pin) low
            comedi.comedi_dio_bitfield2(com, 0, 1, 0, 15)

            # set strobe pin high
            comedi.comedi_dio_bitfield2(com, 0, 1, 1, 16)

            # set strobe pin low
            comedi.comedi_dio_bitfield2(com, 0, 1, 0, 16)

    def set_state(self, condition, **kwargs):
        '''
        Docstring

        Parameters
        ----------

        Returns
        -------
        '''
        self.nidaq.sendMsg(condition)
        super(RelayBlackrock, self).set_state(condition, **kwargs)

    def cleanup(self, database, saveid, **kwargs):
        '''
        Docstring

        Parameters
        ----------

        Returns
        -------
        '''
        super(RelayBlackrock, self).cleanup(database, saveid, **kwargs)
        time.sleep(2)
        for file_ in self.blackrockfiles:
            suffix = file_[-3:]  # e.g., 'nev', 'ns3', etc.
            # database.save_data(file_, "blackrock", saveid, True, False, custom_suffix=suffix)
            database.save_data(file_, "blackrock", saveid, False, False, suffix)

    @classmethod 
    def pre_init(cls, saveid=None, **kwargs):
        if saveid is not None:
            import comedi
            import config
            import time

            com = comedi.comedi_open("/dev/comedi0")
            # set strobe pin low
            comedi.comedi_dio_bitfield2(com, 0, 1, 0, 16)

            # set last data pin ("D15"; 16th pin) high
            comedi.comedi_dio_bitfield2(com, 0, 1, 1, 15)

            # set strobe pin high
            comedi.comedi_dio_bitfield2(com, 0, 1, 1, 16)

            # set strobe pin low
            comedi.comedi_dio_bitfield2(com, 0, 1, 0, 16)

            time.sleep(3)
            super(RelayPlexon, cls).pre_init(saveid=saveid, **kwargs)


class RelayBlackrockByte(RelayBlackrock):
    '''Relays a single byte (0-255) as a row checksum for when a data packet arrives.'''
    
    def init(self):
        '''
        Secondary init function. See riglib.experiment.Experiment.init()
        Prior to starting the task, this 'init' checks that the experiment also uses the SaveHDF feature. 
        '''
    
        if not isinstance(self, SaveHDF):
            raise ValueError("RelayBlackrockByte feature only available with SaveHDF")
        
        super(RelayBlackrockByte, self).init()

    @property
    def ni_out(self):
        '''
        Docstring

        Parameters
        ----------

        Returns
        -------
        '''
        from riglib import nidaq
        return nidaq.SendRowByte


class BlackrockData(object):
    '''Stream Blackrock neural data.'''

    blackrock_channels = None

    def init(self):
        '''
        Secondary init function. See riglib.experiment.Experiment.init()
        Prior to starting the task, this 'init' sets up DataSource objects for streaming from the Blackrock system.
        For LFP streaming, the data is stored as it is received.
        '''

        from riglib import blackrock, source

        if hasattr(self, "_neural_src_type") and hasattr(self, "_neural_src_kwargs") and hasattr(self, "_neural_src_system_type"):
            # for testing only!
            self.neurondata = self._neural_src_type(self._neural_src_system_type, **self._neural_src_kwargs)
        elif 'spike' in self.decoder.extractor_cls.feature_type:  # e.g., 'spike_counts'
            self.neurondata = source.DataSource(blackrock.Spikes, channels=self.blackrock_channels, send_data_to_sink_manager=False)
        elif 'lfp' in self.decoder.extractor_cls.feature_type:  # e.g., 'lfp_power'
            self.neurondata = source.MultiChanDataSource(blackrock.LFP, channels=self.blackrock_channels, send_data_to_sink_manager=True)
        else:
            raise Exception("Unknown extractor class, unable to create data source object!")

        from riglib import sink
        sink_manager = sink.SinkManager.get_instance()
        sink_manager.register(self.neurondata)

        super(BlackrockData, self).init()

    def run(self):
        '''
        Code to execute immediately prior to the beginning of the task FSM executing, or after the FSM has finished running. 
        See riglib.experiment.Experiment.run(). This 'run' method starts the 'neurondata' source before the FSM begins execution
        and stops it after the FSM has completed. 
        '''

        self.neurondata.start()
        try:
            super(BlackrockData, self).run()
        finally:
            self.neurondata.stop()


class BlackrockData(CorticalData):
    @property 
    def sys_module(self):
        return blackrock

class BlackrockBMI(CorticalBMI):
    @property 
    def sys_module(self):
        return blackrock    


# from riglib.bmi.bmi import Decoder
# class BlackrockBMI(BlackrockData, traits.HasTraits):
#     '''
#     Special case of BlackrockData which specifies a subset of channels to stream, i.e., the ones used by the Decoder
#     '''

#     decoder = traits.Instance(Decoder)

#     def init(self):
#         '''
#         Secondary init function. See riglib.experiment.Experiment.init()
#         Prior to starting the task, this 'init' decides which Blackrock channels to stream based on the channels in use by the decoder.
#         '''
#         print "HARDCODING BLACKROCK_CHANNELS -- CHANGE THIS!!"
#         self.blackrock_channels = range(1, 41) #self.decoder.units[:,0]
#         super(BlackrockBMI, self).init()



"""
Link object.

@author: Peter Corke
@copyright: Peter Corke
"""

from numpy import *
from .utility import *
from .transform import *
import copy
import numpy as np
import math
import time


class Link:
    """
    LINK create a new LINK object

    A LINK object holds all information related to a robot link such as
    kinematics of the joint
        - alpha; the link twist angle
        - an; the link length
        - theta; the link rotation angle
        - dn; the link offset
        - sigma; 0 for a revolute joint, non-zero for prismatic
        
    rigid-body inertial parameters
        - I; 3x3 inertia matrix about link COG
        - m; link mass
        - r; link COG wrt link coordinate frame 3x1

    motor and transmission parameters
        - B; link viscous friction (motor referred)
        - Tc; link Coulomb friction 1 element if symmetric, else 2
        - G; gear ratio
        - Jm; inertia (motor referred)

    and miscellaneous
        - qlim; joint limit matrix [lower upper] 2 x 1
        - offset; joint coordinate offset
    Handling the different kinematic conventions is now hidden within the LINK
    object.

    Conceivably all sorts of stuff could live in the LINK object such as
    graphical models of links and so on.

    @see: L{Robot}
    """
    
    LINK_DH = 1
    LINK_MDH = 2

    def __init__(self, theta=0, d=0, a=0, alpha=0, sigma=0, offset=0, convention=LINK_DH):
        """
        L = LINK([theta d a alpha])
        L = LINK([theta d a alpha sigma])
        L = LINK([theta d a alpha sigma offset])
        L = LINK([theta d a alpha], CONVENTION)
        L = LINK([theta d a alpha sigma], CONVENTION)
        L = LINK([theta d a alpha sigma offset], CONVENTION)

        If sigma or offset are not provided they default to zero.  Offset is a
        constant amount added to the joint angle variable before forward kinematics
        and is useful if you want the robot to adopt a 'sensible' pose for zero
        joint angle configuration.

        The optional CONVENTION argument is 'standard' for standard D&H parameters 
        or 'modified' for modified D&H parameters.  If not specified the default
        'standard'.
        """
        self.theta = theta
        self.d = d
        self.a = a
        self.alpha = alpha
        self.sigma = sigma
        self.convention = convention
        self.offset = offset

        # we know nothing about the dynamics
        self.m = None
        self.r = None
        self.v = None
        self.I = None
        self.Jm = None
        self.G = None
        self.B = None
        self.Tc = None
        self.qlim = None

        self._cached_tr = np.mat(np.zeros([4,4]))
        self._cached_tr[-1,-1] = 1
        self.prev_q = -10000

        return None

    def __repr__(self):

        if self.convention == Link.LINK_DH:
            conv = 'std'
        else:
            conv = 'mod'

        if self.sigma == 0:
            jtype = 'R'
        else:
            jtype = 'P'

        if self.d == None:
            return "theta=%f, a=%f, alpha=%f jtype: (%c) conv: (%s)" % (self.theta,
                 self.a, self.alpha, jtype, conv)
        elif self.theta == None:
            return "d=%f, a=%f, alpha=%f, jtype: (%c) conv: (%s)" % (self.d,
                 self.a, self.alpha, jtype, conv)
        else:
            return "theta=%f, d=%f, a=%f, alpha=%f jtype: (%c) conv: (%s)" % (self.theta, 
                 self.d, self.a, self.alpha, jtype, conv)

    # invoked at print
    def __str__(self):
        if self.convention == Link.LINK_DH:
            conv = 'std'
        else:
            conv = 'mod'

        if self.sigma == 0:
            jtype = 'R'
        else:
            jtype = 'P'

        if self.d == None:
            return "theta = %f\ta = %f\talpha = %f\t--\tjtype: %c\tconv: (%s)" % (
                self.theta,  self.a, self.alpha, jtype, conv)
        elif self.theta == None:
            return "d = %f\ta = %f\talpha = %f\t--\tjtype: %c\tconv: (%s)" % (
                self.d, self.a, self.alpha, jtype, conv)
        else:
            return "theta = %f\td=%f\ta = %f\talpha = %f\tjtype: %c\tconv: (%s)" % (
                self.theta, self.d, self.a, self.alpha, jtype, conv)


    def display(self):

        print(self);
        print()

        if self.m != None:
            print("m:", self.m)
        if self.r != None:
            print("r:", self.r)
        if self.I != None:
            print("I:\n", self.I)
        if self.Jm != None:
            print("Jm:", self.Jm)
        if self.B != None:
            print("B:", self.B)
        if self.Tc != None:
            print("Tc:", self.Tc)
        if self.G != None:
            print("G:", self.G)
        if self.qlim != None:
            print("qlim:\n", self.qlim)

    def copy(self):
        """
        Return copy of this Link
        """
        return copy.copy(self);

    def friction(self, qd):
        """
        Compute friction torque for joint rate C{qd}.
        Depending on fields in the Link object viscous and/or Coulomb friction
        are computed.
        
        @type qd: number
        @param qd: joint rate
        @rtype: number
        @return: joint friction torque
        """
        tau = 0.0
        if isinstance(qd, (ndarray, matrix)):
                qd = qd.flatten().T
        if self.B == None:
            self.B = 0
        tau = self.B * qd
        if self.Tc == None:
            self.Tc = mat([0,0])
        tau = tau + (qd > 0) * self.Tc[0,0] + (qd < 0) * self.Tc[0,1]
        return tau
        
    def nofriction(self, all=False):
        """
        Return a copy of the Link object with friction parameters set to zero.
        
        @type all: boolean
        @param all: if True then also zero viscous friction
        @rtype: Link
        @return: Copy of original Link object with zero friction
        @see: L{robot.nofriction}
        """
        
        l2 = self.copy()

        l2.Tc = array([0, 0])
        if all:
            l2.B = 0
        return l2;


    # methods to set kinematic or dynamic parameters

    fields = ["theta", "d", "a", "alpha", "sigma", "offset", "m", "Jm", "G", "B", "convention"];
    
    def __setattr__(self, name, value):
        """
        Set attributes of the Link object
            - theta; scalar
            - d; scalar
            - a; scalar
            - alpha; scalar
            - sigma; scalar
            - offset; scalar
            - m; scalar
            - Jm; scalar
            - G; scalar
            - B; scalar
            - r; 3-vector
            - I; 3x3 matrix, 3-vector or 6-vector
            - Tc; scalar or 2-vector
            - qlim; 2-vector
        
        Inertia, I, can be specified as:
            - 3x3 inertia tensor
            - 3-vector, the diagonal of the inertia tensor
            - 6-vector, the unique elements of the inertia tensor [Ixx Iyy Izz Ixy Iyz Ixz]
            
        Coloumb friction, Tc, can be specifed as:
            - scalar, for the symmetric case when Tc- = Tc+
            - 2-vector, the assymetric case [Tc- Tc+]
            
        Joint angle limits, qlim, is a 2-vector giving the lower and upper limits
        of motion.
        """
    
        if value is None:
            self.__dict__[name] = value
            return

        if name == 'alpha':
            self.__dict__['alpha'] = value
            self.__dict__['sa'] = np.sin(value)
            self.__dict__['ca'] = np.cos(value)
        elif name in self.fields:
            # scalar parameter
            if isinstance(value, (ndarray,matrix)) and value.shape != (1,1):
                raise ValueError("Scalar required")
            if not isinstance(value, (int,float,int32,float64)):
                raise ValueError;
            self.__dict__[name] = value

        elif name == "r":
            r = arg2array(value);
            if len(r) != 3:
                raise ValueError("matrix required")

            self.__dict__[name] = mat(r)
            
        elif name == "I":
            if isinstance(value, matrix) and value.shape == (3,3):
                self.__dict__[name] = value;
            else:
                v = arg2array(value);
                if len(v) == 3:
                    self.__dict__[name] = mat(diag(v))
                elif len(v) == 6:
                    self.__dict__[name] = mat([
                        [v[0],v[3],v[5]],
                        [v[3],v[1],v[4]],
                        [v[5],v[4],v[2]]])
                else:
                    raise ValueError("matrix required");

        elif name == "Tc":
            v = arg2array(value)
            
            if len(v) == 1:
                self.__dict__[name] =  mat([-v[0], v[0]])
            elif len(v) == 2:
                self.__dict__[name] = mat(v)
            else:
                raise ValueError;

        elif name == "qlim":
            v = arg2array(value);
            if len(v) == 2:
                self.__dict__[name] = mat(v);
            else:
                raise ValueError
        elif name == '_cached_tr':
            self.__dict__['_cached_tr'] = value
        elif name == 'prev_q':
            self.__dict__['prev_q'] = value
        else:
            raise NameError("Unknown attribute <%s> of link" % name)


#   LINK.islimit(q) return if limit is exceeded: -1, 0, +1
    def islimit(self,q):
        """
        Check if joint limits exceeded.  Returns
            - -1 if C{q} is less than the lower limit
            - 0 if C{q} is within the limits
            - +1 if C{q} is greater than the high limit
        
        @type q: number
        @param q: Joint coordinate
        @rtype: -1, 0, +1
        @return: joint limit status
        """
        if not self.qlim:
            return 0

        return (q > self.qlim[1,0]) - (q < self.qlim[0,0])

    def tr(self, q, deep_copy=False):
        """
        Compute the transformation matrix for this link.  This is a function
        of kinematic parameters, the kinematic model (DH or MDH) and the joint
        coordinate C{q}.
        
        @type q: number
        @param q: joint coordinatez
        @rtype: homogeneous transformation
        @return: Link transform M{A(q)}
        """
        if hasattr(self, 'offset'):
            q += self.offset

        an = self.a
        dn = self.d
        theta = self.theta

        if self.sigma == 0:
            theta = q   # revolute
        else:
            dn = q      # prismatic

        sa = self.sa #sin(self.alpha)
        ca = self.ca #cos(self.alpha)

        st = math.sin(theta)
        ct = math.cos(theta)

        if self.convention == Link.LINK_DH:
            # standard
            if deep_copy:
                t = np.mat([[ct, -st*ca, st*sa,  an*ct],
                            [st, ct*ca,  -ct*sa, an*st],
                            [0,  sa,     ca,     dn],
                            [0,  0,      0,      1]]);
                return t
            elif q == self.prev_q:
                return self._cached_tr
            else:
                self._cached_tr[0,0] = ct
                self._cached_tr[0,1] = -st*ca
                self._cached_tr[0,2] = st*sa
                self._cached_tr[0,3] = an*ct
                self._cached_tr[1,0] = st
                self._cached_tr[1,1] = ct*ca
                self._cached_tr[1,2] = -ct*sa
                self._cached_tr[1,3] = an*st
                self._cached_tr[2,1] = sa
                self._cached_tr[2,2] = ca
                self._cached_tr[2,3] = dn
                self.prev_q = q
                return self._cached_tr

        else:
            raise NotImplementedError
            # modified
            t =   mat([[ ct,    -st,    0,  an],
                [st*ca, ct*ca,  -sa,    -sa*dn],
                [st*sa, ct*sa,  ca, ca*dn],
                [0, 0,  0,  1]]);

        return t;
    def __eq__(self, other):
        if not isinstance(other, Link):
            return False
        else:
            return self.d == other.d and self.a == other.a and self.alpha == other.alpha and self.sigma == other.sigma and self.convention == other.convention
    



'''
Feature accumulators: The task and the decoder may want to run at
different rates. These modules provide rate-matching
'''
import numpy as np

class FeatureAccumulator(object):
    '''Used only for type-checking'''
    pass

class RectWindowSpikeRateEstimator(FeatureAccumulator):
    '''
    Estimate spike firing rates using a rectangular window
    '''
    def __init__(self, count_max, feature_shape, feature_dtype):
        '''
        Constructor for RectWindowSpikeRateEstimator

        Parameters
        ----------
        count_max : int 
            Number of bins to accumulate in the window. This is somewhat specific
            to rectangular binning
        feature_shape : np.array of shape (n_features, n_timepoints)
            Shape of the extracted features passed to the Decoder on each call
        feature_dtype : np.dtype
            Data type of feature vector. Can be "np.float64" for a vector of numbers
            or something more complicated.

        Returns
        -------
        RectWindowSpikeRateEstimator instance
        '''
        self.count_max = count_max
        self.feature_shape = feature_shape
        self.feature_dtype = feature_dtype
        self.reset()

    def reset(self):
        '''
        Reset the current estimate of the spike rates. Used at the end of the window to clear the estimator for the new window
        '''
        self.est = np.zeros(self.feature_shape, dtype=self.feature_dtype)
        self.count = 0

    def __call__(self, features):
        '''
        Accumulate the current 'features' with the previous estimate

        Parameters
        ----------
        features: np.ndarray of shape self.features_shape 
            self.feature_shape is declared at object creation time

        Returns
        -------
        est: np.ndarray of shape self.features_shape
            Returns current estimate of features. This estimate may or may not be 
            valid depending on when the estimate is checked

        '''
        self.count += 1
        self.est += features
        est = self.est
        decode = False
        if self.count == self.count_max:
            est = self.est.copy()
            self.reset()
            decode = True
        return est, decode

class NullAccumulator(FeatureAccumulator):
    '''
    A null accumulator to use in cases when no accumulation is desired.
    '''
    def __init__(self, count_max):
        '''
        Constructor for NullAccumulator

        Parameters
        ----------
        count_max: int 
            Number of bins to accumulate in the window. This is somewhat specific
            to rectangular binning

        Returns
        -------
        NullAccumulator instance
        '''
        self.count_max = count_max
        self.reset()

    def reset(self):
        '''
        Reset the counter
        '''
        self.count = 0

    def __call__(self, features):
        '''
        Accumulate the current 'features' with the previous estimate

        Parameters
        ----------
        features: np.ndarray of shape self.features_shape 
            self.feature_shape is declared at object creation time

        Returns
        -------
        est: np.ndarray of shape self.features_shape
            Returns current estimate of features. This estimate may or may not be 
            valid depending on when the estimate is checked
        '''
        self.count += 1
        decode = False
        if self.count == self.count_max:
            self.reset()
            decode = True
        return features, decode


'''
Features for interacting with neuropixels
'''

import datetime
import numpy as np
from riglib.experiment import traits
from riglib import experiment
from riglib.open_ephys import OpenEphysHTTPServer

class RecordNeuropixels(traits.HasTraits):

    neuropixel_port1_chamber = traits.String("", desc="the channel position (e.g. LM1)")
    neuropixel_port1_drive_type = traits.String("", desc="the name of the hole layout")
    neuropixel_port1_site = traits.Int(0, desc="the hole number of the penetration")
    neuropixel_port1_depth = traits.Float(0, desc="the recording depth in um")
    neuropxiel_port2_chamber = traits.String("", desc="the channel position (e.g. LM1)")
    neuropixel_port2_drive_type = traits.String("", desc="the name of the hole layout")
    neuropixel_port2_site = traits.Int(0, desc="the hole number of the penetration")
    neuropixel_port2_depth = traits.Float(0, desc="the recording depth in um")
    gui = OpenEphysHTTPServer('10.155.205.108', timeout=0.5)
    
    def cleanup(self,database, saveid, gui=gui, **kwargs):
        super().cleanup(database, saveid, gui=gui, **kwargs)
        try:
            gui.acquire()
        except Exception as e:
            print(e)
            print('\n\ncould not stop OpenEphys recording. Please manually stop the recording\n\n')
        
    @classmethod
    def pre_init(cls, saveid=None, subject_name=None, gui=gui,**kwargs):
        cls.openephys_status = 'IDLE'
        prepend_text = str(datetime.date.today())
        filename = f'_Neuropixel_{subject_name}_te'
        append_text = str(saveid) if saveid else 'Test'
        parent_dir = 'E://Neuropixel_data'

        gui.set_parent_dir(parent_dir)
        gui.set_prepend_text(prepend_text)
        gui.set_base_text(filename)
        gui.set_append_text(append_text)

        if saveid is not None:
            try:
                gui.record()
                cls.openephys_status = gui.status()
            except Exception as e:
                print(e)
                print('\n\ncould not start OpenEphys recording\n\n')
        else:
            try:
                gui.acquire()
                cls.openephys_status = gui.status()
            except Exception as e:
                print(e)
                print('\n\ncould not start OpenEphys acquisition\n\n')
        print(f'Open Ephys status : {gui.status()}')
  
        if hasattr(super(), 'pre_init'):
            super().pre_init(saveid=saveid,gui=gui,**kwargs)

    def run(self):
        if not self.openephys_status in ["ACQUIRE", "RECORD"]:
            import io
            self.terminated_in_error = True
            self.termination_err = io.StringIO()
            self.termination_err.write(self.openephys_status)
            self.termination_err.seek(0)
            self.state = None
        try:
            super().run()
        except Exception as e:
            print(e)



'''
Django config file mostly auto-generated when a django project is created.
See https://docs.djangoproject.com/en/dev/intro/tutorial01/ for an introduction
on how to customize this file test
'''
import os, glob, re
import glob
import re
import socket
import json
from django.core.exceptions import ImproperlyConfigured
from config.rig_defaults import db as rig_defaults

BASE_DIR = os.path.abspath(os.path.dirname(__file__))
CONFIG_DIR = os.path.join(os.path.expanduser("~"), '.config', 'bmi3d')
HOSTNAME = socket.gethostname()
DEFAULT_AUTO_FIELD = 'django.db.models.AutoField'

# Add databases
def get_sqlite3_databases():
    dbs = dict()
    db_files = glob.glob(os.path.join(BASE_DIR, '*.sql'))
    for db in db_files:
        db_name_re = re.match('db(.*?).sql', os.path.basename(db))
        db_name = db_name_re.group(1)
        if db_name.startswith('_'):
            db_name = db_name[1:]
        elif db_name == "":
            db_name = "local"
        else:
            # unrecognized db name pattern
            print("Unrecognized database file: ", db)
            continue

        dbs[db_name] = {
            'ENGINE': 'django.db.backends.sqlite3', # Add 'postgresql_psycopg2', 'postgresql', 'mysql', 'sqlite3' or'oracle'.
            'NAME': db,                      # Or path to database file if using sqlite3.
            'USER': '',                      # Not used with sqlite3.
            'PASSWORD': '',                  # Not used with sqlite3.
            'HOST': '',                      # Set to empty string for localhost. Not used with sqlite3.
            'PORT': '',                      # Set to empty string for default. Not used with sqlite3.        
        }

    return dbs

def get_secret(setting):
    """Get secret setting or fail with ImproperlyConfigured"""
    try:
        with open(os.path.join(CONFIG_DIR, 'secrets.json')) as secrets_file:
            secrets = json.load(secrets_file)
        return secrets[setting]
    except KeyError:
        raise ImproperlyConfigured("Set the {} setting".format(setting))

def get_mysql_database(dbname):
    return {
        'ENGINE': 'django.db.backends.mysql',
        'NAME': dbname,
        'HOST': get_secret("DB_HOST"),
        'PORT': get_secret("DB_PORT"),
        'USER': get_secret("DB_USER"),
        'PASSWORD': get_secret("DB_PASSWORD")
    }

# DATABASES = { 'default': { 'ENGINE': 'django.db.backends.sqlite3', 'NAME': 'db.sql', } }

DATABASES = get_sqlite3_databases() # local databases

dbnames = rig_defaults['secret_dbnames']
for dbname in dbnames:
    DATABASES[dbname] = get_mysql_database(dbname)

if rig_defaults['enable_celery']:
    CELERY_BROKER_URL = f'amqp://{get_secret("AMQP_USER")}:{get_secret("AMQP_PASSWORD")}@{get_secret("AMQP_HOST")}:{get_secret("AMQP_PORT")}//'

default_db = rig_defaults['default_db']
if default_db not in DATABASES:
    print('No databases configured. Setting test_aopy as default database.')
    default_db = 'test_aopy'
if os.environ.get('BMI3D_TEST_DATABASE'):
    DATABASES['default'] = DATABASES['test']
else:
    DATABASES['default'] = DATABASES[default_db]

# Django settings for db project.
DEBUG = True
TEMPLATE_DEBUG = DEBUG

ADMINS = (
    # ('Your Name', 'your_email@domain.com'),
)

MANAGERS = ADMINS

# Local time zone for this installation. Choices can be found here:
# http://en.wikipedia.org/wiki/List_of_tz_zones_by_name
# although not all choices may be available on all operating systems.
# On Unix systems, a value of None will cause Django to use the same
# timezone as the operating system.
# If running in a Windows environment this must be set to the same as your
# system time zone.
TIME_ZONE = 'America/Los_Angeles'

# Language code for this installation. All choices can be found here:
# http://www.i18nguy.com/unicode/language-identifiers.html
LANGUAGE_CODE = 'en-us'

SITE_ID = 1

# If you set this to False, Django will make some optimizations so as not
# to load the internationalization machinery.
USE_I18N = True

# If you set this to False, Django will not format dates, numbers and
# calendars according to the current locale
USE_L10N = True

# Absolute filesystem path to the directory that will hold user-uploaded files.
# Example: "/home/media/media.lawrence.com/"
MEDIA_ROOT = ''

# URL that handles the media served from MEDIA_ROOT. Make sure to use a
# trailing slash if there is a path component (optional in other cases).
# Examples: "http://media.lawrence.com", "http://example.com/media/"
MEDIA_URL = ''

# URL prefix for admin media -- CSS, JavaScript and images. Make sure to use a
# trailing slash.
# Examples: "http://foo.com/media/", "/media/".
ADMIN_MEDIA_PREFIX = '/media/'

# Make this unique, and don't share it with anybody.
SECRET_KEY = 'i$k&ifgz&_=m+!(1dm*9g3a1v6ue6#10_r!j!y6g^oj=1ha_z!'

# List of callables that know how to import templates from various sources.
TEMPLATE_LOADERS = (
    'django.template.loaders.filesystem.Loader',
    'django.template.loaders.app_directories.Loader',
#     'django.template.loaders.eggs.Loader',
)

MIDDLEWARE_CLASSES = (
    'django.middleware.common.CommonMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
)

MIDDLEWARE = [
    'django.middleware.security.SecurityMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',
]


ROOT_URLCONF = 'db.urls'

TEMPLATE_DIRS = (
    os.path.join(BASE_DIR, "html", "templates"),
)

STATIC_URL = "/static/"

STATICFILES_DIRS = (
    os.path.join(BASE_DIR, "html", "static"),
    "/usr/share/pyshared/django/contrib/"
)
ADMIN_MEDIA_PREFIX = '/static/admin/'
INSTALLED_APPS = (
    'db.tracker',
    'django.contrib.auth',
    'django.contrib.admin',    
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.messages',
    'django.contrib.staticfiles',
)

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.jinja2.Jinja2',
        'DIRS': [],
        'APP_DIRS': True,
        'OPTIONS': {
            'environment': 'db.tracker.jinja2.environment'
        },
    },
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [],
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                'django.template.context_processors.debug',
                'django.template.context_processors.request',
                'django.contrib.auth.context_processors.auth',
                'django.contrib.messages.context_processors.messages',
            ],
        },
    },
]

LOGGING_CONFIG = None
LOGGING = {
    'version': 1,
    'disable_existing_loggers': False,
    'handlers': {
        'console': {
            'class': 'logging.StreamHandler',
        },
    },
    'root': {
        'handlers': ['console'],
        'level': 'WARNING',
    },
}

APPEND_SLASH=False
ALLOWED_HOSTS = ['*'] #['127.0.0.1', 'localhost', "testserver"]





from .tracker import models

def hide_empty_task_entries(hide=False):
	tes_empty_report = models.TaskEntry.objects.filter(report='')

	# filter out task entries which somehow have data files associated with them.
	tes_to_hide = []
	for te in tes_empty_report:
		assoc_dfs = models.DataFile.objects.filter(entry_id=te.id)
		if len(assoc_dfs) == 0:
			tes_to_hide.append(te)

	# filter out TaskEntry records which are already hidden
	tes_to_hide = [te for te in tes_to_hide if te.visible]

	if hide:
		print("hiding %d entries" % len(tes_to_hide))
		for te in tes_to_hide:
			te.visible = False
			te.save()

def calc_distribution_of_n_trials():
	n_trials = dict()
	visible_tes = models.TaskEntry.objects.filter(visible=True)
	for te in visible_tes:
		print(te.id)
		try:
			n_trials[te.id] = te.offline_report()['Total trials']
		except:
			pass
	return n_trials


'''Needs docs'''

import numpy as np
from OpenGL.GL import *

from .render import Renderer
from .fbo import FBOrender, FBO
from ..textures import Texture

class SSAO(FBOrender):
    def __init__(self, *args, **kwargs):
        super(SSAO, self).__init__(*args, **kwargs)
        self.sf = 3
        w, h = self.size[0] / self.sf, self.size[1] / self.sf
        
        self.normdepth = FBO(["color0", "depth"], size=(w,h))
        self.ping = FBO(['color0'], size=(w,h))
        self.pong = FBO(["color0"], size=(w,h))

        self.add_shader("fsquad", GL_VERTEX_SHADER, "fsquad.v.glsl")
        self.add_shader("ssao_pass1", GL_FRAGMENT_SHADER, "ssao_pass1.f.glsl")
        self.add_shader("ssao_pass2", GL_FRAGMENT_SHADER, "ssao_pass2.f.glsl")
        self.add_shader("ssao_pass3", GL_FRAGMENT_SHADER, "ssao_pass3.f.glsl", "phong.f.glsl")
        self.add_shader("hblur", GL_FRAGMENT_SHADER, "hblur.f.glsl")
        self.add_shader("vblur", GL_FRAGMENT_SHADER, "vblur.f.glsl")

        #override the default shader with this passthru + ssao_pass1 to store depth
        self.add_program("ssao_pass1", ("passthru", "ssao_pass1"))
        self.add_program("ssao_pass2", ("fsquad", "ssao_pass2"))
        self.add_program("hblur", ("fsquad", "hblur"))
        self.add_program("vblur", ("fsquad", "vblur"))
        self.add_program("ssao_pass3", ("passthru", "ssao_pass3"))

        randtex = np.random.rand(3, w, h)
        randtex /= randtex.sum(0)
        self.rnm = Texture(randtex.T, wrap_x=GL_REPEAT, wrap_y=GL_REPEAT, 
            magfilter=GL_NEAREST, minfilter=GL_NEAREST)
        self.rnm.init()

        self.clips = args[2], args[3]

    def draw(self, root, **kwargs):
        #First, draw the whole damned scene, but only read the normals and depth into ssao
        glPushAttrib(GL_VIEWPORT_BIT)
        glViewport( 0,0, self.size[0]/self.sf, self.size[1]/self.sf)
        self.draw_to_fbo(self.normdepth, root, shader="ssao_pass1", **kwargs)
        
        #Now, do the actual ssao calculations, and draw it into ping
        self.draw_fsquad_to_fbo(self.pong, "ssao_pass2", rnm=self.rnm,
            normalMap=self.normdepth['color0'], depthMap=self.normdepth['depth'],
            nearclip=self.clips[0], farclip=self.clips[1] )
        
        #Blur the textures
        self.draw_fsquad_to_fbo(self.ping, "hblur", tex=self.pong['color0'], blur=1./(self.size[0]/self.sf))
        self.draw_fsquad_to_fbo(self.pong, "vblur", tex=self.ping['color0'], blur=1./(self.size[0]/self.sf))
        
        glPopAttrib()
        #Actually draw the final image to the screen
        win = glGetIntegerv(GL_VIEWPORT)
        #Why is this call necessary at all?!
        glViewport(*win)
        
        super(SSAO, self).draw(root, shader="ssao_pass3", shadow=self.pong['color0'], 
            window=[float(i) for i in win], **kwargs)
        
        #self.draw_done()
    
    def clear(self):
        self.normdepth.clear()
        self.ping.clear()
        self.pong.clear()
    
    def draw_done(self):
        super(SSAO, self).draw_done()
        self.clear()



# Generated by Django 2.2.1 on 2020-10-14 08:54

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('tracker', '0005_taskentry_sw_version'),
    ]

    operations = [
        migrations.AddField(
            model_name='datafile',
            name='backup_status',
            field=models.CharField(blank=True, max_length=256, null=True),
        ),
    ]



from __future__ import absolute_import, unicode_literals
import os
from celery import Celery

# set the default Django settings module for the 'celery' program.
from ..boot_django import boot_django
boot_django()

app = Celery('tracker')

# Using a string here means the worker doesn't have to serialize
# the configuration object to child processes.
# - namespace='CELERY' means all celery-related configuration keys
#   should have a `CELERY_` prefix.
#app.config_from_object('django.conf:settings')
app.config_from_object('django.conf:settings', namespace='CELERY')

# Load task modules from all registered Django app configs.
from django.conf import settings
app.autodiscover_tasks(settings.INSTALLED_APPS)


@app.task(bind=True)
def debug_task(self):
    print('Request: {0!r}'.format(self.request))



'''
Laser delivery features
'''

import time
import traceback
from riglib.experiment import traits
from riglib.gpio import ArduinoGPIO, DigitalWave
import numpy as np
from riglib.lasers.qwalor_laser import QwalorLaserSerial, QwalorLaserSwitched

class CrystaLaser(traits.HasTraits):
    ''' Adds an arduino-controlled crystalaser to self.lasers.'''

    # laser_serial_port = traits.Str(desc="Serial port used to communicate with arduino")
    crystalaser_pin = traits.Int(12, desc="Pin number for laser")

    def __init__(self, *args, **kwargs):
        self.lasers = []
        super().__init__(*args, **kwargs)

    def init(self, *args, **kwargs):
        laser = ArduinoGPIO('/dev/crystalaser')
        laser.port = self.crystalaser_pin
        laser.name = 'crystalaser'
        laser.set_power = lambda x: None
        self.lasers.append(laser)
        super().init(*args, **kwargs)

class QwalorLaser(traits.HasTraits):
    '''
    Enable a single QWALOR laser channel
    '''

    # laser_serial_port = traits.Str(desc="Serial port used to communicate with arduino")
    qwalor_channel = traits.Int(1, desc="Laser channel (1-red, 2-blue, 3-green, 4-blue)")
    qwalor_trigger_dch = traits.Int(9, desc="Digital channel (0-index) recording laser trigger")
    qwalor_sensor_ach = traits.Int(16, desc="Analog channel (0-index) recording laser power")
    stimulation_site = traits.String("", desc="Where was the laser stimulation?")
    stimulation_channel_mapping_file = traits.String("", desc="Name of channel mapping excel file")
    stimulation_chamber = traits.String("", desc="Name of the stimulation chamber (e.g. LM1)")
    stimulation_drive_type = traits.String("", desc="Type of stimulation drive (e.g. Opto32)")
    qwalor_peak_power = traits.Float(1.0, desc="Maximum power (in watts) measured at the fiber output")
    qwalor_laser_version = traits.Int(1, desc="Laser protocol version number")

    hidden_traits = ['qwalor_laser_version']

    def __init__(self, *args, **kwargs):
        self.lasers = []
        super().__init__(*args, **kwargs)

    def init(self, *args, **kwargs):

        # Attempt to open the laser connection, but fail gracefully if it is unavailable
        try:
            laser = QwalorLaserSerial(self.qwalor_channel)
            self.qwalor_laser_status = 'ok'
            
            # Add the laser to the list of available lasers
            laser.port = laser.trigger_pin
            laser.name = 'qwalor_laser'
            self.lasers.append(laser)
            time.sleep(3) # some extra time to make sure the lasers are initialized

        except Exception as e:
            traceback.print_exc()
            self.qwalor_laser_status = 'Couldn\'t connect to laser modulator, make it is turned on!'
            
        super().init(*args, **kwargs)

    def run(self):
        if not self.qwalor_laser_status == 'ok':
            import io
            self.terminated_in_error = True
            self.termination_err = io.StringIO()
            self.termination_err.write(self.qwalor_laser_status)
            self.termination_err.seek(0)
            self.state = None
        super().run()

class SwitchedQwalorLaser(QwalorLaser):

    n_channels = 16 # number of channels on the optical switch
    stimulation_site = traits.Array(value=[-1]*n_channels, 
                                    desc="Where was the laser stimulation at each connected channel. -1 means the channel is not connected. 0 means consult the channel mapping file")
    qwalor_switch_rdy_dch = traits.Int(60, desc="Digital channel (0-index) recording the READY signal from the optical switch")
    qwalor_switch_data_dch = traits.Array(value=[56, 57, 58, 59], desc="Digital channels (0-index) recording the DATA signals from the optical switch")

    hidden_traits = ['qwalor_switch_rdy_dch', 'qwalor_switch_data_dch']

    def init(self, *args, **kwargs):

        # Attempt to open the laser connection, but fail gracefully if it is unavailable
        try:
            for ch in range(1, len(self.stimulation_site)+1):

                # Skip channels that are not connected
                if self.stimulation_site[ch-1] == -1:
                    continue

                # Add the laser to the list of available lasers
                laser = QwalorLaserSwitched(laser_channel=self.qwalor_channel, switch_channel=ch-1) # 0-indexed in the switch
                laser.port = laser.trigger_pin
                laser.name = f'qwalor_laser_switch_ch{ch}'
                self.lasers.append(laser)
                
            time.sleep(3) # some extra time to make sure the lasers are initialized
            self.qwalor_laser_status = 'ok'

        except Exception as e:
            traceback.print_exc()
            self.qwalor_laser_status = 'Couldn\'t connect to laser modulator, make it is turned on!'
            
        super().init(*args, **kwargs)


class MultiQwalorLaser(traits.HasTraits):
    '''
    Enable multiple QWALOR laser channels
    '''

    qwalor_ch1_enable = traits.Bool(False, desc="Laser channel 1-red")
    qwalor_ch2_enable = traits.Bool(False, desc="Laser channel 2-blue")
    qwalor_ch3_enable = traits.Bool(False, desc="Laser channel 3-green")
    qwalor_ch4_enable = traits.Bool(False, desc="Laser channel 4-blue")
    qwalor_ch1_trigger_dch = traits.Int(8, desc="Digital channel (0-index) recording laser ch1 trigger")
    qwalor_ch2_trigger_dch = traits.Int(9, desc="Digital channel (0-index) recording laser ch2 trigger")
    qwalor_ch3_trigger_dch = traits.Int(10, desc="Digital channel (0-index) recording laser ch3 trigger")
    qwalor_ch4_trigger_dch = traits.Int(11, desc="Digital channel (0-index) recording laser ch4 trigger")
    qwalor_ch1_sensor_ach = traits.Int(15, desc="Analog channel (0-index) recording laser ch1 power")
    qwalor_ch2_sensor_ach = traits.Int(16, desc="Analog channel (0-index) recording laser ch2 power")
    qwalor_ch3_sensor_ach = traits.Int(17, desc="Analog channel (0-index) recording laser ch3 power")
    qwalor_ch4_sensor_ach = traits.Int(18, desc="Analog channel (0-index) recording laser ch4 power")
    stimulation_site_ch1 = traits.String("", desc="Where was the ch1 laser stimulation?")
    stimulation_site_ch2 = traits.String("", desc="Where was the ch2 laser stimulation?")
    stimulation_site_ch3 = traits.String("", desc="Where was the ch3 laser stimulation?")
    stimulation_site_ch4 = traits.String("", desc="Where was the ch4 laser stimulation?")
    stimulation_channel_mapping_file = traits.String("", desc="Name of channel mapping excel file")
    stimulation_chamber = traits.String("", desc="Name of the stimulation chamber (e.g. LM1)")
    stimulation_drive_type = traits.String("", desc="Type of stimulation drive (e.g. Opto32)")
    qwalor_ch1_peak_power = traits.Float(1.0, desc="Maximum power (in watts) measured at the fiber output from ch1")
    qwalor_ch2_peak_power = traits.Float(1.0, desc="Maximum power (in watts) measured at the fiber output from ch2")
    qwalor_ch3_peak_power = traits.Float(1.0, desc="Maximum power (in watts) measured at the fiber output from ch3")
    qwalor_ch4_peak_power = traits.Float(1.0, desc="Maximum power (in watts) measured at the fiber output from ch4")
    qwalor_laser_version = traits.Int(1, desc="Laser protocol version number")

    hidden_traits = ['qwalor_ch1_sensor_ach', 'qwalor_ch2_sensor_ach', 'qwalor_ch3_sensor_ach', 'qwalor_ch4_sensor_ach',
                     'qwalor_ch1_trigger_dch', 'qwalor_ch2_trigger_dch', 'qwalor_ch3_trigger_dch', 'qwalor_ch4_trigger_dch',
                     'qwalor_laser_version']

    def __init__(self, *args, **kwargs):
        self.lasers = []
        super().__init__(*args, **kwargs)

    def init(self, *args, **kwargs):

        # Attempt to open the laser connections, but fail gracefully if it is unavailable
        try:
            if self.qwalor_ch1_enable:
                laser = QwalorLaserSerial(1)
                laser.port = laser.trigger_pin
                laser.name = 'qwalor_laser_ch1'
                self.lasers.append(laser)
            
            if self.qwalor_ch2_enable:
                laser = QwalorLaserSerial(2)
                laser.port = laser.trigger_pin
                laser.name = 'qwalor_laser_ch2'
                self.lasers.append(laser)

            if self.qwalor_ch3_enable:
                laser = QwalorLaserSerial(3)
                laser.port = laser.trigger_pin
                laser.name = 'qwalor_laser_ch3'
                self.lasers.append(laser)

            if self.qwalor_ch4_enable:
                laser = QwalorLaserSerial(4)
                laser.port = laser.trigger_pin
                laser.name = 'qwalor_laser_ch4'
                self.lasers.append(laser)

            time.sleep(3) # some extra time to make sure the lasers are initialized
            self.qwalor_laser_status = 'ok'
            
        except Exception as e:
            traceback.print_exc()
            self.qwalor_laser_status = 'Couldn\'t connect to laser modulator, make it is turned on!'
            
        super().init(*args, **kwargs)

    def run(self):
        if not self.qwalor_laser_status == 'ok':
            import io
            self.terminated_in_error = True
            self.termination_err = io.StringIO()
            self.termination_err.write(self.qwalor_laser_status)
            self.termination_err.seek(0)
            self.state = None
        super().run()

class LaserState(traits.HasTraits):
    '''
    Trigger lasers to stimulate at the beginning of a given state. Make sure the state duration is 
    longer than the total stimulation duration otherwise the stim may extend into other states.
    '''

    laser_trigger_state = traits.String("wait", desc="State machine state that triggers laser")
    laser_end_state = traits.String("", desc="Optional state machine state that ends laser stimulation")
    laser_stims_per_trial = traits.Int(1, desc="Number of stimulations per laser per trial")
    laser_power = traits.List([1.,], desc="Laser power (between 0 and 1) for each active laser")
    laser_pulse_width = traits.List([0.005,], desc="List of possible pulse widths in seconds")
    laser_poisson_mu = traits.Float(0.5, desc="Mean duration between laser stimulations (s)")
    laser_min_isi = traits.Float(0.01, desc="Minimum duration between laser stimulation pulses (s)")
    laser_max_stim_time = traits.Float(0., desc="Maximum time after state begins after which no more "
        "pulses are generated (s). Time of 0. is unlimited.")
    laser_rand_start = traits.Bool(True, desc="Randomize the timing of the first laser pulse")
    laser_rand_order = traits.Bool(True, desc="Randomize the order of the laser channels")
    lasers_per_trial = traits.Int(0, desc="Number of lasers to use per trial. 0 means use all available lasers.")

    def __init__(self, *args, **kwargs):
        self.laser_threads = []
        super().__init__(*args, **kwargs)
    
    def init(self, *args, **kwargs):            
        super().init(*args, **kwargs)            
        
        for idx in range(len(self.lasers)):
            laser = self.lasers[idx]
            power = self.laser_power[idx]
            laser.set_power(power)

    def run(self):
        if not (hasattr(self, 'lasers') and len(self.lasers) > 0):
            import io
            self.terminated_in_error = True
            self.termination_err = io.StringIO()
            self.termination_err.write("No laser feature enabled, cannot init LaserState")
            self.termination_err.seek(0)
            self.state = None
        if len(self.lasers) != len(self.laser_power):
            import io
            self.terminated_in_error = True
            self.termination_err = io.StringIO()
            self.termination_err.write("Incorrect length laser power attribute. Please match to number of lasers.")
            self.termination_err.seek(0)
            self.state = None
        super().run() 

    def start_state(self, state):
        super().start_state(state)
        if state == self.laser_end_state:
            for wave in self.laser_waves:
                wave.stop()
            return

        if state != self.laser_trigger_state:
            return
        
        if self.laser_rand_start:
            wait_time = np.random.exponential(self.laser_poisson_mu)
        else:
            wait_time = 0
        order = np.arange(len(self.lasers)).astype(int)
        if self.laser_rand_order:
            np.random.shuffle(order)
            if self.lasers_per_trial > 0:
                order = order[:self.lasers_per_trial] # only use a random subset of the lasers
        self.laser_waves = []
        for idx in order:
            laser = self.lasers[idx]

            width_idx = np.random.choice(len(self.laser_pulse_width))
            width = self.laser_pulse_width[width_idx]

            for n in range(self.laser_stims_per_trial):

                # Skip the pulses if we've reached the stim time limit
                if (self.laser_max_stim_time > 0. and 
                    wait_time+width > self.laser_max_stim_time):
                    continue

                # Trigger digital wave
                wave = DigitalWave(laser, mask=1<<laser.port)
                wave.set_edges([wait_time, wait_time+width], True)
                wave.start()
                self.laser_waves.append(wave)

                # Make the next pulse come after a delay
                delay = max(np.random.exponential(self.laser_poisson_mu), self.laser_min_isi)
                wait_time += delay



"""
Python toolbox utility and helper functions.


@author: Peter Corke
@copyright: Peter Corke
"""

from numpy import *


def ishomog(tr):
    """
    True if C{tr} is a 4x4 homogeneous transform.
    
    @note: Only the dimensions are tested, not whether the rotation submatrix
    is orthonormal.
    
    @rtype: boolean
    """
    
    return tr.shape == (4,4)



def isrot(r):
    """
    True if C{tr} is a 3x3 matrix.
    
    @note: Only the dimensions are tested, not whether the matrix
    is orthonormal.
    
    @rtype: boolean
    """
    return r.shape == (3,3)


def isvec(v, l=3):
    """
    True if C{tr} is an l-vector.  
    
    @param v: object to test
    @type l: integer
    @param l: length of vector (default 3)
   
    @rtype: boolean
    """
    return v.shape == (l,1) or v.shape == (1,l) or v.shape == (l,)


def numcols(m):
    """
    Number of columns in a matrix.
    
    @type m: matrix
    @return: the number of columns in the matrix.
    return m.shape[1];
    """
    return m.shape[1];
    
def numrows(m):
    """
    Number of rows in a matrix.
    
    @type m: matrix
    @return: the number of rows in the matrix.
    return m.shape[1];
    """
    return m.shape[0];

################ vector operations

def unit(v):
    """
    Unit vector.
    
    @type v: vector
    @rtype: vector
    @return: unit-vector parallel to C{v}
    """
    return mat(v / linalg.norm(v))
    
def crossp(v1, v2):
    """
    Vector cross product.
    
    @note Differs from L{numpy.cross} in that vectors can be row or
    column.
    
    @type v1: 3-vector
    @type v2: 3-vector
    @rtype: 3-vector
    @return: Cross product M{v1 x v2}
    """
    v1=mat(v1)
    v2=mat(v2)
    v1=v1.reshape(3,1)
    v2=v2.reshape(3,1)
    v = matrix(zeros( (3,1) ))
    v[0] = v1[1]*v2[2] - v1[2]*v2[1]
    v[1] = v1[2]*v2[0] - v1[0]*v2[2]
    v[2] = v1[0]*v2[1] - v1[1]*v2[0]
    return v

################ misc support functions

def arg2array(arg):
    """
    Convert a 1-dimensional argument that is either a list, array or matrix to an 
    array.
    
    Useful for functions where the argument might be in any of these formats:::
            func(a)
            func(1,2,3)
            
            def func(*args):
                if len(args) == 1:
                    v = arg2array(arg[0]);
                elif len(args) == 3:
                    v = arg2array(args);
             .
             .
             .
    
    @rtype: array
    @return: Array equivalent to C{arg}.
    """
    if isinstance(arg, (matrix, ndarray)):
        s = arg.shape;
        if len(s) == 1:
            return array(arg);
        if min(s) == 1:
            return array(arg).flatten();

    elif isinstance(arg, list):
        return array(arg);

    elif isinstance(arg, (int, float, float32, float64)):
        return array([arg]);
        
    raise ValueError;
        


import traceback;

def error(s):
    """
    Common error handler.  Display the error string, execute a traceback then raise
    an execption to return to the interactive prompt.
    """
    print('Robotics toolbox error:', s)

    #traceback.print_exc();
    raise ValueError
    
    
if __name__ == "__main__":
    print(arg2array(1))
    print(arg2array(1.0)); 
    print(arg2array( mat([1,2,3,4]) ))
    print(arg2array( mat([1,2,3,4]).T ))
    print(arg2array( array([1,2,3]) ));
    print(arg2array( array([1,2,3]).T ));
    print(arg2array( [1,2,3]));

    
